"""Shared pytest fixtures for snapshot testing."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from syrupy.extensions.json import JSONSnapshotExtension
from syrupy.extensions.single_file import SingleFileSnapshotExtension, WriteMode

if TYPE_CHECKING:
    from syrupy.assertion import SnapshotAssertion


class YamlSnapshotExtension(SingleFileSnapshotExtension):
    """Snapshot extension that persists plain-text YAML files."""

    file_extension = "yaml"
    _write_mode = WriteMode.TEXT


@pytest.fixture
def snapshot_json(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    """JSON-formatted snapshot assertions."""
    return snapshot.with_defaults(extension_class=JSONSnapshotExtension)


@pytest.fixture
def snapshot_yaml(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    """YAML-formatted snapshot assertions."""
    return snapshot.with_defaults(extension_class=YamlSnapshotExtension)



"""Color output utilities for vcspull CLI."""

from __future__ import annotations

import os
import sys
from enum import Enum

from colorama import Fore, Style


class ColorMode(Enum):
    """Color output modes."""

    AUTO = "auto"
    ALWAYS = "always"
    NEVER = "never"


class Colors:
    """Semantic color constants and utilities."""

    # Semantic colors
    SUCCESS = Fore.GREEN  # Success, additions, up-to-date
    WARNING = Fore.YELLOW  # Warnings, changes needed, behind remote
    ERROR = Fore.RED  # Errors, deletions, conflicts
    INFO = Fore.CYAN  # Information, paths, URLs
    HIGHLIGHT = Fore.MAGENTA  # Workspace roots, important labels
    MUTED = Fore.BLUE  # Subdued info, bullets
    RESET = Style.RESET_ALL

    def __init__(self, mode: ColorMode = ColorMode.AUTO) -> None:
        """Initialize color manager.

        Parameters
        ----------
        mode : ColorMode
            Color mode to use (auto, always, never)
        """
        self.mode = mode
        self._enabled = self._should_enable_color()

    def _should_enable_color(self) -> bool:
        """Determine if color should be enabled.

        Returns
        -------
        bool
            True if colors should be enabled
        """
        # Respect NO_COLOR environment variable
        if os.environ.get("NO_COLOR"):
            return False

        if self.mode == ColorMode.NEVER:
            return False
        if self.mode == ColorMode.ALWAYS:
            return True

        # AUTO mode: check if stdout is a TTY
        return sys.stdout.isatty()

    def colorize(self, text: str, color: str) -> str:
        """Apply color to text if colors are enabled.

        Parameters
        ----------
        text : str
            Text to colorize
        color : str
            Color code (e.g., Fore.GREEN)

        Returns
        -------
        str
            Colorized text if enabled, plain text otherwise
        """
        if self._enabled:
            return f"{color}{text}{self.RESET}"
        return text

    def success(self, text: str) -> str:
        """Format text as success (green)."""
        return self.colorize(text, self.SUCCESS)

    def warning(self, text: str) -> str:
        """Format text as warning (yellow)."""
        return self.colorize(text, self.WARNING)

    def error(self, text: str) -> str:
        """Format text as error (red)."""
        return self.colorize(text, self.ERROR)

    def info(self, text: str) -> str:
        """Format text as info (cyan)."""
        return self.colorize(text, self.INFO)

    def highlight(self, text: str) -> str:
        """Format text as highlighted (magenta)."""
        return self.colorize(text, self.HIGHLIGHT)

    def muted(self, text: str) -> str:
        """Format text as muted (blue)."""
        return self.colorize(text, self.MUTED)


def get_color_mode(color_arg: str | None = None) -> ColorMode:
    """Determine color mode from argument.

    Parameters
    ----------
    color_arg : str | None
        Color mode argument (auto, always, never)

    Returns
    -------
    ColorMode
        The determined color mode
    """
    if color_arg is None:
        return ColorMode.AUTO

    try:
        return ColorMode(color_arg.lower())
    except ValueError:
        return ColorMode.AUTO



"""Tests for config reader utilities."""

from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

from vcspull._internal.config_reader import DuplicateAwareConfigReader

if TYPE_CHECKING:
    import pathlib


def _write(tmp_path: pathlib.Path, name: str, content: str) -> pathlib.Path:
    file_path = tmp_path / name
    file_path.write_text(content, encoding="utf-8")
    return file_path


def test_duplicate_aware_reader_records_yaml_duplicates(tmp_path: pathlib.Path) -> None:
    """YAML duplicate workspace roots are captured without data loss."""
    yaml_content = textwrap.dedent(
        """\
        ~/study/python/:
          repo1:
            repo: git+https://example.com/repo1.git
        ~/study/python/:
          repo2:
            repo: git+https://example.com/repo2.git
        """,
    )
    config_path = _write(tmp_path, "config.yaml", yaml_content)

    reader = DuplicateAwareConfigReader.from_file(config_path)

    assert reader.content == {
        "~/study/python/": {
            "repo2": {"repo": "git+https://example.com/repo2.git"},
        },
    }
    assert "~/study/python/" in reader.duplicate_sections
    captured = reader.duplicate_sections["~/study/python/"]
    assert len(captured) == 2
    assert captured[0] == {
        "repo1": {"repo": "git+https://example.com/repo1.git"},
    }
    assert captured[1] == {
        "repo2": {"repo": "git+https://example.com/repo2.git"},
    }


def test_duplicate_aware_reader_handles_yaml_without_duplicates(
    tmp_path: pathlib.Path,
) -> None:
    """YAML without duplicate keys reports an empty duplicate map."""
    yaml_content = textwrap.dedent(
        """\
        ~/code/:
          repo:
            repo: git+https://example.com/repo.git
        """,
    )
    config_path = _write(tmp_path, "single.yaml", yaml_content)

    reader = DuplicateAwareConfigReader.from_file(config_path)

    assert reader.content == {
        "~/code/": {
            "repo": {"repo": "git+https://example.com/repo.git"},
        },
    }
    assert reader.duplicate_sections == {}


def test_duplicate_aware_reader_passes_through_json(tmp_path: pathlib.Path) -> None:
    """JSON configs remain supported and duplicates remain empty."""
    json_content = textwrap.dedent(
        """\
        {
          "~/code/": {
            "repo": {"repo": "git+https://example.com/repo.git"}
          }
        }
        """,
    )
    config_path = _write(tmp_path, "config.json", json_content)

    reader = DuplicateAwareConfigReader.from_file(config_path)

    assert reader.content == {
        "~/code/": {
            "repo": {"repo": "git+https://example.com/repo.git"},
        },
    }
    assert reader.duplicate_sections == {}


def test_duplicate_aware_reader_preserves_top_level_item_order(
    tmp_path: pathlib.Path,
) -> None:
    """Loader should expose ordered top-level items so duplicates can be replayed."""
    yaml_content = textwrap.dedent(
        """\
        ~/study/python/:
          Flexget:
            repo: git+https://github.com/Flexget/Flexget.git
        ~/study/python/:
          bubbles:
            repo: git+https://github.com/Stiivi/bubbles.git
        ~/study/python/:
          cubes:
            repo: git+https://github.com/Stiivi/cubes.git
        """,
    )
    config_path = _write(tmp_path, "ordered.yaml", yaml_content)

    reader = DuplicateAwareConfigReader.from_file(config_path)

    items = reader.top_level_items
    assert [key for key, _ in items] == [
        "~/study/python/",
        "~/study/python/",
        "~/study/python/",
    ]
    assert items[0][1] == {
        "Flexget": {"repo": "git+https://github.com/Flexget/Flexget.git"},
    }
    assert items[1][1] == {
        "bubbles": {"repo": "git+https://github.com/Stiivi/bubbles.git"},
    }
    assert items[2][1] == {
        "cubes": {"repo": "git+https://github.com/Stiivi/cubes.git"},
    }



"""Tests for vcspull configuration format."""

from __future__ import annotations

import logging
import textwrap
import typing as t

import pytest

from vcspull import config

if t.TYPE_CHECKING:
    import pathlib

    from vcspull.types import ConfigDict, RawConfigDict


class LoadYAMLFn(t.Protocol):
    """Typing for load_yaml pytest fixture."""

    def __call__(
        self,
        content: str,
        path: str = "randomdir",
        filename: str = "randomfilename.yaml",
    ) -> tuple[pathlib.Path, list[t.Any | pathlib.Path], list[ConfigDict]]:
        """Callable function type signature for load_yaml pytest fixture."""
        ...


@pytest.fixture
def load_yaml(tmp_path: pathlib.Path) -> LoadYAMLFn:
    """Return a yaml loading function that uses temporary directory path."""

    def fn(
        content: str,
        path: str = "randomdir",
        filename: str = "randomfilename.yaml",
    ) -> tuple[pathlib.Path, list[pathlib.Path], list[ConfigDict]]:
        """Return vcspull configurations and write out config to temp directory."""
        dir_ = tmp_path / path
        dir_.mkdir()
        config_ = dir_ / filename
        config_.write_text(content, encoding="utf-8")

        configs = config.find_config_files(path=dir_)
        repos = config.load_configs(configs, cwd=dir_)
        return dir_, configs, repos

    return fn


def test_simple_format(load_yaml: LoadYAMLFn) -> None:
    """Test simple configuration YAML file for vcspull."""
    path, _, repos = load_yaml(
        """
vcspull:
  libvcs: git+https://github.com/vcs-python/libvcs
   """,
    )

    assert len(repos) == 1
    repo = repos[0]

    assert path / "vcspull" == repo["path"].parent
    assert path / "vcspull" / "libvcs" == repo["path"]


def test_relative_dir(load_yaml: LoadYAMLFn) -> None:
    """Test configuration files for vcspull support relative directories."""
    path, _, repos = load_yaml(
        """
./relativedir:
  docutils: svn+http://svn.code.sf.net/p/docutils/code/trunk
   """,
    )

    config_files = config.find_config_files(path=path)
    repos = config.load_configs(config_files, path)

    assert len(repos) == 1
    repo = repos[0]

    assert path / "relativedir" == repo["path"].parent
    assert path / "relativedir" / "docutils" == repo["path"]


class ExtractWorkspaceFixture(t.NamedTuple):
    """Fixture capturing workspace root injection scenarios."""

    test_id: str
    raw_config: dict[str, dict[str, str | dict[str, str]]]
    expected_roots: dict[str, str]


EXTRACT_WORKSPACE_FIXTURES: list[ExtractWorkspaceFixture] = [
    ExtractWorkspaceFixture(
        test_id="tilde-workspace",
        raw_config={
            "~/code/": {
                "alpha": {"repo": "git+https://example.com/alpha.git"},
            },
        },
        expected_roots={"alpha": "~/code/"},
    ),
    ExtractWorkspaceFixture(
        test_id="relative-workspace",
        raw_config={
            "./projects": {
                "beta": "git+https://example.com/beta.git",
            },
        },
        expected_roots={"beta": "./projects"},
    ),
]


@pytest.mark.parametrize(
    list(ExtractWorkspaceFixture._fields),
    EXTRACT_WORKSPACE_FIXTURES,
    ids=[fixture.test_id for fixture in EXTRACT_WORKSPACE_FIXTURES],
)
def test_extract_repos_injects_workspace_root(
    test_id: str,
    raw_config: dict[str, dict[str, str | dict[str, str]]],
    expected_roots: dict[str, str],
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Ensure extract_repos assigns workspace_root consistently."""
    import pathlib as pl

    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.chdir(tmp_path)

    typed_raw_config = t.cast("RawConfigDict", raw_config)
    repos = config.extract_repos(typed_raw_config, cwd=tmp_path)

    assert len(repos) == len(expected_roots)

    for repo in repos:
        name = repo["name"]
        expected_root = expected_roots[name]
        assert repo["workspace_root"] == expected_root
        expected_path = config.expand_dir(pl.Path(expected_root), cwd=tmp_path) / name
        assert repo["path"] == expected_path


def _write_duplicate_config(tmp_path: pathlib.Path) -> pathlib.Path:
    config_path = tmp_path / "config.yaml"
    config_path.write_text(
        textwrap.dedent(
            """\
            ~/workspace/:
              alpha:
                repo: git+https://example.com/alpha.git
            ~/workspace/:
              beta:
                repo: git+https://example.com/beta.git
            """,
        ),
        encoding="utf-8",
    )
    return config_path


def test_load_configs_merges_duplicate_workspace_roots(
    tmp_path: pathlib.Path,
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Duplicate workspace roots are merged to keep every repository."""
    monkeypatch.setenv("HOME", str(tmp_path))
    caplog.set_level(logging.INFO, logger="vcspull.config")

    config_path = _write_duplicate_config(tmp_path)

    repos = config.load_configs([config_path], cwd=tmp_path)

    repo_names = {repo["name"] for repo in repos}
    assert repo_names == {"alpha", "beta"}

    merged_messages = [message for message in caplog.messages if "merged" in message]
    assert merged_messages, "Expected a merge log entry for duplicate roots"


def test_load_configs_can_skip_merging_duplicates(
    tmp_path: pathlib.Path,
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """The merge step can be skipped while still warning about duplicates."""
    monkeypatch.setenv("HOME", str(tmp_path))
    caplog.set_level(logging.WARNING, logger="vcspull.config")

    config_path = _write_duplicate_config(tmp_path)

    repos = config.load_configs(
        [config_path],
        cwd=tmp_path,
        merge_duplicates=False,
    )

    repo_names = {repo["name"] for repo in repos}
    assert repo_names == {"beta"}

    warning_messages = [
        message for message in caplog.messages if "duplicate" in message
    ]
    assert warning_messages, "Expected a warning about duplicate workspace roots"



"""Tests for vcspull logging utilities."""

from __future__ import annotations

import logging
import typing as t

import pytest
from colorama import Fore

from vcspull.log import (
    LEVEL_COLORS,
    DebugLogFormatter,
    LogFormatter,
    RepoFilter,
    RepoLogFormatter,
    SimpleLogFormatter,
    get_cli_logger_names,
    setup_logger,
)

if t.TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


@pytest.fixture(autouse=True)
def cleanup_loggers() -> t.Iterator[None]:
    """Clean up logger configuration after each test."""
    managed_loggers = [
        "",
        "vcspull",
        *get_cli_logger_names(include_self=True),
        "libvcs",
        "test_logger",
    ]

    for logger_name in managed_loggers:
        logger = logging.getLogger(logger_name)
        logger.handlers.clear()
        logger.setLevel(logging.NOTSET)
        logger.propagate = True

    yield

    for logger_name in managed_loggers:
        logger = logging.getLogger(logger_name)
        logger.handlers.clear()
        logger.setLevel(logging.NOTSET)
        logger.propagate = True


def test_level_colors_defined() -> None:
    """Test that all standard log levels have color mappings."""
    expected_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    for level in expected_levels:
        assert level in LEVEL_COLORS


def test_level_color_values() -> None:
    """Test that color values are correct colorama constants."""
    assert LEVEL_COLORS["DEBUG"] == Fore.BLUE
    assert LEVEL_COLORS["INFO"] == Fore.GREEN
    assert LEVEL_COLORS["WARNING"] == Fore.YELLOW
    assert LEVEL_COLORS["ERROR"] == Fore.RED
    assert LEVEL_COLORS["CRITICAL"] == Fore.RED


class LogFormatterTestCase(t.NamedTuple):
    """Test case for log formatter behavior."""

    test_id: str
    level: str
    message: str
    logger_name: str
    expected_contains: list[str]
    expected_not_contains: list[str] = []  # noqa: RUF012


LOG_FORMATTER_TEST_CASES: list[LogFormatterTestCase] = [
    LogFormatterTestCase(
        test_id="info_level",
        level="INFO",
        message="test info message",
        logger_name="test.logger",
        expected_contains=["(INFO)", "test info message", "test.logger"],
    ),
    LogFormatterTestCase(
        test_id="debug_level",
        level="DEBUG",
        message="debug information",
        logger_name="debug.logger",
        expected_contains=["(DEBUG)", "debug information", "debug.logger"],
    ),
    LogFormatterTestCase(
        test_id="warning_level",
        level="WARNING",
        message="warning message",
        logger_name="warn.logger",
        expected_contains=["(WARNING)", "warning message", "warn.logger"],
    ),
    LogFormatterTestCase(
        test_id="error_level",
        level="ERROR",
        message="error occurred",
        logger_name="error.logger",
        expected_contains=["(ERROR)", "error occurred", "error.logger"],
    ),
]


def test_log_formatter_template_includes_required_elements() -> None:
    """Test that template includes all required formatting elements."""
    formatter = LogFormatter()
    record = logging.LogRecord(
        name="test.logger",
        level=logging.INFO,
        pathname="/test/path.py",
        lineno=42,
        msg="test message",
        args=(),
        exc_info=None,
    )

    template = formatter.template(record)

    # Should include levelname, asctime, and name placeholders
    assert "%(levelname)" in template
    assert "%(asctime)s" in template
    assert "%(name)s" in template


def test_log_formatter_basic_message() -> None:
    """Test formatting a basic log message."""
    formatter = LogFormatter()
    record = logging.LogRecord(
        name="test.logger",
        level=logging.INFO,
        pathname="/test/path.py",
        lineno=42,
        msg="test message",
        args=(),
        exc_info=None,
    )

    result = formatter.format(record)

    assert "test message" in result
    assert "test.logger" in result
    assert "(INFO)" in result


def test_log_formatter_handles_newlines() -> None:
    """Test that multiline messages are properly indented."""
    formatter = LogFormatter()
    record = logging.LogRecord(
        name="test.logger",
        level=logging.INFO,
        pathname="/test/path.py",
        lineno=42,
        msg="line 1\nline 2\nline 3",
        args=(),
        exc_info=None,
    )

    result = formatter.format(record)

    # Newlines should be replaced with newline + indent
    assert "\n    line 2" in result
    assert "\n    line 3" in result


def test_log_formatter_handles_bad_message() -> None:
    """Test formatter handles malformed log messages gracefully."""
    formatter = LogFormatter()

    # Create a record that will cause getMessage() to fail
    record = logging.LogRecord(
        name="test.logger",
        level=logging.INFO,
        pathname="/test/path.py",
        lineno=42,
        msg="bad format %s %s",  # Two placeholders
        args=("only_one_arg",),  # But only one argument
        exc_info=None,
    )

    result = formatter.format(record)

    assert "Bad message" in result


@pytest.mark.parametrize(
    list(LogFormatterTestCase._fields),
    LOG_FORMATTER_TEST_CASES,
    ids=[test.test_id for test in LOG_FORMATTER_TEST_CASES],
)
def test_log_formatter_levels(
    test_id: str,
    level: str,
    message: str,
    logger_name: str,
    expected_contains: list[str],
    expected_not_contains: list[str],
) -> None:
    """Test formatter with different log levels."""
    formatter = LogFormatter()
    level_int = getattr(logging, level)

    record = logging.LogRecord(
        name=logger_name,
        level=level_int,
        pathname="/test/path.py",
        lineno=42,
        msg=message,
        args=(),
        exc_info=None,
    )

    result = formatter.format(record)

    for expected in expected_contains:
        assert expected in result

    for not_expected in expected_not_contains:
        assert not_expected not in result


def test_debug_formatter_template_includes_debug_elements() -> None:
    """Debug template should include module and function info."""
    formatter = DebugLogFormatter()
    record = logging.LogRecord(
        name="test.logger",
        level=logging.INFO,
        pathname="/test/module.py",
        lineno=42,
        msg="test message",
        args=(),
        exc_info=None,
    )
    record.module = "test_module"
    record.funcName = "test_function"

    template = formatter.template(record)

    assert "%(module)s.%(funcName)s()" in template
    assert "%(lineno)d" in template


def test_debug_formatter_output_includes_debug_info() -> None:
    """Formatting should include debug metadata when level is DEBUG."""
    formatter = DebugLogFormatter()
    record = logging.LogRecord(
        name="test.logger",
        level=logging.DEBUG,
        pathname="/test/module.py",
        lineno=123,
        msg="debug message",
        args=(),
        exc_info=None,
    )
    record.module = "test_module"
    record.funcName = "test_function"

    result = formatter.format(record)

    assert "debug message" in result
    assert "test_module.test_function()" in result
    assert "123" in result
    assert "(D)" in result


def test_simple_formatter_returns_only_message() -> None:
    """Simple formatter should return only the message."""
    formatter = SimpleLogFormatter()
    record = logging.LogRecord(
        name="test.logger",
        level=logging.INFO,
        pathname="/test/path.py",
        lineno=42,
        msg="simple message",
        args=(),
        exc_info=None,
    )

    result = formatter.format(record)

    assert result == "simple message"


def test_simple_formatter_handles_arguments() -> None:
    """Simple formatter should expand arguments."""
    formatter = SimpleLogFormatter()
    record = logging.LogRecord(
        name="test.logger",
        level=logging.INFO,
        pathname="/test/path.py",
        lineno=42,
        msg="message with %s and %d",
        args=("text", 42),
        exc_info=None,
    )

    result = formatter.format(record)

    assert result == "message with text and 42"


def test_simple_formatter_excludes_metadata() -> None:
    """Simple formatter should exclude metadata such as logger name."""
    formatter = SimpleLogFormatter()
    record = logging.LogRecord(
        name="very.long.logger.name",
        level=logging.WARNING,
        pathname="/very/long/path/to/module.py",
        lineno=999,
        msg="clean message",
        args=(),
        exc_info=None,
    )

    result = formatter.format(record)

    assert "very.long.logger.name" not in result
    assert "WARNING" not in result
    assert "/very/long/path" not in result
    assert "999" not in result
    assert result == "clean message"


def test_repo_formatter_template_references_repo_info() -> None:
    """Repo template should inject bin_name and keyword values."""
    formatter = RepoLogFormatter()
    record = logging.LogRecord(
        name="libvcs.sync.git",
        level=logging.INFO,
        pathname="/libvcs/sync/git.py",
        lineno=42,
        msg="git operation",
        args=(),
        exc_info=None,
    )
    record.bin_name = "git"
    record.keyword = "clone"
    record.message = "git operation"  # RepoLogFormatter expects this

    template = formatter.template(record)

    assert "git" in template
    assert "clone" in template


def test_repo_formatter_formats_message() -> None:
    """Formatted repo log should include bin_name, keyword, and message."""
    formatter = RepoLogFormatter()
    record = logging.LogRecord(
        name="libvcs.sync.git",
        level=logging.INFO,
        pathname="/libvcs/sync/git.py",
        lineno=42,
        msg="Cloning repository",
        args=(),
        exc_info=None,
    )
    record.bin_name = "git"
    record.keyword = "clone"

    result = formatter.format(record)

    assert "git" in result
    assert "clone" in result
    assert "Cloning repository" in result


def test_repo_filter_accepts_repo_records() -> None:
    """Filter should accept records with keyword attribute."""
    repo_filter = RepoFilter()
    record = logging.LogRecord(
        name="libvcs.sync.git",
        level=logging.INFO,
        pathname="/libvcs/sync/git.py",
        lineno=42,
        msg="repo message",
        args=(),
        exc_info=None,
    )
    record.keyword = "clone"

    assert repo_filter.filter(record) is True


def test_repo_filter_rejects_non_repo_records() -> None:
    """Filter should reject records without keyword attribute."""
    repo_filter = RepoFilter()
    record = logging.LogRecord(
        name="regular.logger",
        level=logging.INFO,
        pathname="/regular/module.py",
        lineno=42,
        msg="regular message",
        args=(),
        exc_info=None,
    )

    assert repo_filter.filter(record) is False


def test_repo_filter_handles_keyword_in_dict() -> None:
    """Filter should honor keyword set directly on record dict."""
    repo_filter = RepoFilter()
    record = logging.LogRecord(
        name="libvcs.sync.git",
        level=logging.INFO,
        pathname="/libvcs/sync/git.py",
        lineno=42,
        msg="repo message",
        args=(),
        exc_info=None,
    )
    record.__dict__["keyword"] = "pull"

    assert repo_filter.filter(record) is True


def test_get_cli_logger_names_includes_base() -> None:
    """Helper returns expected CLI modules including base package."""
    names = get_cli_logger_names(include_self=True)
    expected = [
        "vcspull.cli",
        "vcspull.cli._colors",
        "vcspull.cli._output",
        "vcspull.cli._workspaces",
        "vcspull.cli.add",
        "vcspull.cli.discover",
        "vcspull.cli.fmt",
        "vcspull.cli.list",
        "vcspull.cli.status",
        "vcspull.cli.sync",
    ]
    assert names == expected


def test_get_cli_logger_names_without_base() -> None:
    """Helper omits base package when include_self is False."""
    names = get_cli_logger_names(include_self=False)
    assert "vcspull.cli" not in names
    assert all(name.startswith("vcspull.cli.") for name in names)


def test_setup_logger_default_behavior(caplog: LogCaptureFixture) -> None:
    """setup_logger should configure vcspull logger once at INFO level."""
    test_logger = logging.getLogger("test_logger")
    test_logger.handlers.clear()

    setup_logger(test_logger, level="INFO")

    vcspull_logger = logging.getLogger("vcspull")
    assert len(vcspull_logger.handlers) > 0
    assert vcspull_logger.propagate is True
    handler = vcspull_logger.handlers[0]
    assert isinstance(handler.formatter, SimpleLogFormatter)


def test_setup_logger_custom_level(caplog: LogCaptureFixture) -> None:
    """setup_logger should honor a DEBUG log level."""
    setup_logger(level="DEBUG")

    vcspull_logger = logging.getLogger("vcspull")
    assert vcspull_logger.level == logging.DEBUG
    handler = vcspull_logger.handlers[0]
    assert isinstance(handler.formatter, DebugLogFormatter)


def test_setup_logger_initializes_vcspull_logger(caplog: LogCaptureFixture) -> None:
    """Vcspull logger should exist with a simple formatter."""
    setup_logger(level="INFO")

    vcspull_logger = logging.getLogger("vcspull")
    assert len(vcspull_logger.handlers) > 0
    assert vcspull_logger.propagate is True
    handler = vcspull_logger.handlers[0]
    assert isinstance(handler.formatter, SimpleLogFormatter)


def test_setup_logger_leaves_cli_loggers_propagating(caplog: LogCaptureFixture) -> None:
    """CLI loggers should propagate without their own handlers."""
    setup_logger(level="INFO")

    for logger_name in [
        "vcspull.cli.add",
        "vcspull.cli.sync",
    ]:
        logger = logging.getLogger(logger_name)
        assert logger.propagate is True
        assert len(logger.handlers) == 0


def test_setup_logger_configures_libvcs_logger(caplog: LogCaptureFixture) -> None:
    """Libvcs logger should receive RepoLogFormatter handler."""
    setup_logger(level="INFO")

    libvcs_logger = logging.getLogger("libvcs")
    assert len(libvcs_logger.handlers) > 0
    assert libvcs_logger.propagate is True
    handler = libvcs_logger.handlers[0]
    assert isinstance(handler.formatter, RepoLogFormatter)


def test_setup_logger_avoids_duplicate_handlers(caplog: LogCaptureFixture) -> None:
    """setup_logger should not add handlers to the provided logger."""
    test_logger = logging.getLogger("test_logger")
    test_logger.handlers.clear()

    setup_logger(test_logger, level="INFO")
    assert len(test_logger.handlers) == 0

    setup_logger(test_logger, level="INFO")
    assert len(test_logger.handlers) == 0


def test_setup_logger_with_none_creates_root_logger(
    caplog: LogCaptureFixture,
) -> None:
    """When no logger is provided, vcspull logger should be configured."""
    setup_logger(log=None, level="WARNING")

    vcspull_logger = logging.getLogger("vcspull")
    assert len(vcspull_logger.handlers) > 0
    assert vcspull_logger.level == logging.WARNING


def test_simple_formatter_integration(caplog: LogCaptureFixture) -> None:
    """Simple formatter should integrate with logging.Logger."""
    logger = logging.getLogger("test_simple")
    logger.handlers.clear()

    handler = logging.StreamHandler()
    handler.setFormatter(SimpleLogFormatter())
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    logger.info("clean message")

    assert "clean message" in caplog.text


def test_debug_formatter_integration(caplog: LogCaptureFixture) -> None:
    """Debug formatter should output debug metadata."""
    logger = logging.getLogger("test_debug")
    logger.handlers.clear()

    handler = logging.StreamHandler()
    handler.setFormatter(DebugLogFormatter())
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)

    logger.debug("debug message")

    assert "debug message" in caplog.text


def test_repo_filter_integration(caplog: LogCaptureFixture) -> None:
    """RepoFilter should pass repo logs and ignore others."""
    logger = logging.getLogger("test_repo")
    logger.handlers.clear()
    logger.propagate = False

    handler = logging.StreamHandler()
    handler.setFormatter(RepoLogFormatter())
    handler.addFilter(RepoFilter())
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    record = logging.LogRecord(
        name="test_repo",
        level=logging.INFO,
        pathname="test.py",
        lineno=1,
        msg="repo operation",
        args=(),
        exc_info=None,
    )
    record.bin_name = "git"
    record.keyword = "clone"
    record.message = "repo operation"  # RepoLogFormatter expects this

    logger.handle(record)

    logger.info("regular message")

    repo_filter = RepoFilter()
    assert repo_filter.filter(record) is True

    regular_record = logging.LogRecord(
        name="test_repo",
        level=logging.INFO,
        pathname="test.py",
        lineno=1,
        msg="regular message",
        args=(),
        exc_info=None,
    )
    assert repo_filter.filter(regular_record) is False



"""CLI utilities for vcspull."""

from __future__ import annotations

import argparse
import logging
import pathlib
import textwrap
import typing as t
from typing import overload

from libvcs.__about__ import __version__ as libvcs_version

from vcspull.__about__ import __version__
from vcspull.log import setup_logger

from ._formatter import VcspullHelpFormatter
from .add import add_repo, create_add_subparser, handle_add_command
from .discover import create_discover_subparser, discover_repos
from .fmt import create_fmt_subparser, format_config_file
from .list import create_list_subparser, list_repos
from .status import create_status_subparser, status_repos
from .sync import create_sync_subparser, sync

log = logging.getLogger(__name__)


def build_description(
    intro: str,
    example_blocks: t.Sequence[tuple[str | None, t.Sequence[str]]],
) -> str:
    """Assemble help text with optional example sections."""
    sections: list[str] = []
    intro_text = textwrap.dedent(intro).strip()
    if intro_text:
        sections.append(intro_text)

    for heading, commands in example_blocks:
        if not commands:
            continue
        title = "examples:" if heading is None else f"{heading} examples:"
        lines = [title]
        lines.extend(f"  {command}" for command in commands)
        sections.append("\n".join(lines))

    return "\n\n".join(sections)


CLI_DESCRIPTION = build_description(
    """
    Manage multiple VCS repositories from a single configuration file.
    """,
    (
        (
            "sync",
            [
                'vcspull sync "*"',
                'vcspull sync "django-*"',
                'vcspull sync --dry-run "*"',
                'vcspull sync -f ./myrepos.yaml "*"',
                "vcspull sync -w ~/code myproject",
            ],
        ),
        (
            "list",
            [
                "vcspull list",
                'vcspull list "django-*"',
                "vcspull list --tree",
                "vcspull list --json",
            ],
        ),
        (
            "add",
            [
                "vcspull add mylib https://github.com/example/mylib.git",
                "vcspull add mylib URL -w ~/code",
                "vcspull add mylib URL --dry-run",
            ],
        ),
        (
            "discover",
            [
                "vcspull discover ~/code",
                "vcspull discover ~/code --recursive --yes",
                "vcspull discover ~/code -w ~/projects --dry-run",
            ],
        ),
        (
            "fmt",
            [
                "vcspull fmt",
                "vcspull fmt -f ./myrepos.yaml",
                "vcspull fmt --write",
                "vcspull fmt --all",
            ],
        ),
    ),
)

SYNC_DESCRIPTION = build_description(
    """
    Synchronize VCS repositories.
    """,
    (
        (
            None,
            [
                'vcspull sync "*"',
                'vcspull sync "django-*"',
                'vcspull sync --dry-run "*"',
                'vcspull sync -f ./myrepos.yaml "*"',
                "vcspull sync -w ~/code myproject",
            ],
        ),
    ),
)

LIST_DESCRIPTION = build_description(
    """
    List configured repositories.
    """,
    (
        (
            None,
            [
                "vcspull list",
                'vcspull list "django-*"',
                "vcspull list --tree",
                "vcspull list --json",
            ],
        ),
    ),
)

STATUS_DESCRIPTION = build_description(
    """
    Check status of repositories.
    """,
    (
        (
            None,
            [
                "vcspull status",
                'vcspull status "django-*"',
                "vcspull status --detailed",
                "vcspull status --json",
            ],
        ),
    ),
)

ADD_DESCRIPTION = build_description(
    """
    Add a single repository to the configuration.
    """,
    (
        (
            None,
            [
                "vcspull add mylib https://github.com/example/mylib.git",
                "vcspull add mylib URL -w ~/code",
                "vcspull add mylib URL --dry-run",
            ],
        ),
    ),
)

DISCOVER_DESCRIPTION = build_description(
    """
    Discover and add repositories from filesystem.

    Scans a directory for git repositories and adds them to the configuration.
    """,
    (
        (
            None,
            [
                "vcspull discover ~/code",
                "vcspull discover ~/code --recursive --yes",
                "vcspull discover ~/code -w ~/projects --dry-run",
            ],
        ),
    ),
)

FMT_DESCRIPTION = build_description(
    """
    Format vcspull configuration files for consistency.

    Normalizes repository entries, sorts sections, and can write changes
    back to disk or format all discovered configuration files.
    """,
    (
        (
            None,
            [
                "vcspull fmt",
                "vcspull fmt -f ./myrepos.yaml",
                "vcspull fmt --write",
                "vcspull fmt --all",
            ],
        ),
    ),
)


@overload
def create_parser(
    return_subparsers: t.Literal[True],
) -> tuple[argparse.ArgumentParser, t.Any]: ...


@overload
def create_parser(return_subparsers: t.Literal[False]) -> argparse.ArgumentParser: ...


def create_parser(
    return_subparsers: bool = False,
) -> argparse.ArgumentParser | tuple[argparse.ArgumentParser, t.Any]:
    """Create CLI argument parser for vcspull."""
    parser = argparse.ArgumentParser(
        prog="vcspull",
        formatter_class=VcspullHelpFormatter,
        description=CLI_DESCRIPTION,
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {__version__}, libvcs {libvcs_version}",
    )
    parser.add_argument(
        "--log-level",
        metavar="level",
        action="store",
        default="INFO",
        help="log level (debug, info, warning, error, critical)",
    )

    subparsers = parser.add_subparsers(dest="subparser_name")

    # Sync command
    sync_parser = subparsers.add_parser(
        "sync",
        help="synchronize repositories",
        formatter_class=VcspullHelpFormatter,
        description=SYNC_DESCRIPTION,
    )
    create_sync_subparser(sync_parser)

    # List command
    list_parser = subparsers.add_parser(
        "list",
        help="list configured repositories",
        formatter_class=VcspullHelpFormatter,
        description=LIST_DESCRIPTION,
    )
    create_list_subparser(list_parser)

    # Status command
    status_parser = subparsers.add_parser(
        "status",
        help="check repository status",
        formatter_class=VcspullHelpFormatter,
        description=STATUS_DESCRIPTION,
    )
    create_status_subparser(status_parser)

    # Add command
    add_parser = subparsers.add_parser(
        "add",
        help="add a single repository",
        formatter_class=VcspullHelpFormatter,
        description=ADD_DESCRIPTION,
    )
    create_add_subparser(add_parser)

    # Discover command
    discover_parser = subparsers.add_parser(
        "discover",
        help="discover repositories from filesystem",
        formatter_class=VcspullHelpFormatter,
        description=DISCOVER_DESCRIPTION,
    )
    create_discover_subparser(discover_parser)

    # Fmt command
    fmt_parser = subparsers.add_parser(
        "fmt",
        help="format configuration files",
        formatter_class=VcspullHelpFormatter,
        description=FMT_DESCRIPTION,
    )
    create_fmt_subparser(fmt_parser)

    if return_subparsers:
        # Return all parsers needed by cli() function
        return parser, (
            sync_parser,
            list_parser,
            status_parser,
            add_parser,
            discover_parser,
            fmt_parser,
        )
    return parser


def cli(_args: list[str] | None = None) -> None:
    """CLI entry point for vcspull."""
    parser, subparsers = create_parser(return_subparsers=True)
    (
        sync_parser,
        _list_parser,
        _status_parser,
        _add_parser,
        _discover_parser,
        _fmt_parser,
    ) = subparsers
    args = parser.parse_args(_args)

    setup_logger(log=log, level=args.log_level.upper())

    if args.subparser_name is None:
        parser.print_help()
        return

    if args.subparser_name == "sync":
        sync(
            repo_patterns=args.repo_patterns,
            config=pathlib.Path(args.config) if args.config else None,
            workspace_root=getattr(args, "workspace_root", None),
            dry_run=getattr(args, "dry_run", False),
            output_json=getattr(args, "output_json", False),
            output_ndjson=getattr(args, "output_ndjson", False),
            color=getattr(args, "color", "auto"),
            exit_on_error=args.exit_on_error,
            show_unchanged=getattr(args, "show_unchanged", False),
            summary_only=getattr(args, "summary_only", False),
            long_view=getattr(args, "long_view", False),
            relative_paths=getattr(args, "relative_paths", False),
            fetch=getattr(args, "fetch", False),
            offline=getattr(args, "offline", False),
            verbosity=getattr(args, "verbosity", 0),
            parser=sync_parser,
        )
    elif args.subparser_name == "list":
        list_repos(
            repo_patterns=args.repo_patterns,
            config_path=pathlib.Path(args.config) if args.config else None,
            workspace_root=getattr(args, "workspace_root", None),
            tree=args.tree,
            output_json=args.output_json,
            output_ndjson=args.output_ndjson,
            color=args.color,
        )
    elif args.subparser_name == "status":
        status_repos(
            repo_patterns=args.repo_patterns,
            config_path=pathlib.Path(args.config) if args.config else None,
            workspace_root=getattr(args, "workspace_root", None),
            detailed=args.detailed,
            output_json=args.output_json,
            output_ndjson=args.output_ndjson,
            color=args.color,
            concurrent=not getattr(args, "no_concurrent", False),
            max_concurrent=getattr(args, "max_concurrent", None),
        )
    elif args.subparser_name == "add":
        handle_add_command(args)
    elif args.subparser_name == "discover":
        discover_repos(
            scan_dir_str=args.scan_dir,
            config_file_path_str=args.config,
            recursive=args.recursive,
            workspace_root_override=args.workspace_root_path,
            yes=args.yes,
            dry_run=args.dry_run,
            merge_duplicates=args.merge_duplicates,
        )
    elif args.subparser_name == "fmt":
        format_config_file(
            args.config,
            args.write,
            args.all,
            merge_roots=args.merge_roots,
        )



"""Tests for vcspull list command."""

from __future__ import annotations

import json
import typing as t

import pytest
import yaml

from vcspull.cli.list import list_repos

if t.TYPE_CHECKING:
    import pathlib

    from _pytest.monkeypatch import MonkeyPatch


def create_test_config(config_path: pathlib.Path, repos: dict[str, t.Any]) -> None:
    """Create a test config file."""
    with config_path.open("w", encoding="utf-8") as f:
        yaml.dump(repos, f)


class ListReposFixture(t.NamedTuple):
    """Fixture for list repos test cases."""

    test_id: str
    config_data: dict[str, t.Any]
    patterns: list[str]
    tree: bool
    output_json: bool
    output_ndjson: bool
    workspace_filter: str | None
    expected_repo_names: list[str]


LIST_REPOS_FIXTURES: list[ListReposFixture] = [
    ListReposFixture(
        test_id="list-all-repos",
        config_data={
            "~/code/": {
                "flask": {"repo": "git+https://github.com/pallets/flask.git"},
                "django": {"repo": "git+https://github.com/django/django.git"},
            },
        },
        patterns=[],
        tree=False,
        output_json=False,
        output_ndjson=False,
        workspace_filter=None,
        expected_repo_names=["flask", "django"],
    ),
    ListReposFixture(
        test_id="list-with-pattern",
        config_data={
            "~/code/": {
                "flask": {"repo": "git+https://github.com/pallets/flask.git"},
                "django": {"repo": "git+https://github.com/django/django.git"},
                "requests": {"repo": "git+https://github.com/psf/requests.git"},
            },
        },
        patterns=["fla*"],
        tree=False,
        output_json=False,
        output_ndjson=False,
        workspace_filter=None,
        expected_repo_names=["flask"],
    ),
    ListReposFixture(
        test_id="list-json-output",
        config_data={
            "~/code/": {
                "flask": {"repo": "git+https://github.com/pallets/flask.git"},
            },
        },
        patterns=[],
        tree=False,
        output_json=True,
        output_ndjson=False,
        workspace_filter=None,
        expected_repo_names=["flask"],
    ),
    ListReposFixture(
        test_id="list-ndjson-output",
        config_data={
            "~/code/": {
                "flask": {"repo": "git+https://github.com/pallets/flask.git"},
            },
        },
        patterns=[],
        tree=False,
        output_json=False,
        output_ndjson=True,
        workspace_filter=None,
        expected_repo_names=["flask"],
    ),
    ListReposFixture(
        test_id="list-workspace-filter",
        config_data={
            "~/code/": {
                "flask": {"repo": "git+https://github.com/pallets/flask.git"},
            },
            "~/work/": {
                "internal": {"repo": "git+https://github.com/user/internal.git"},
            },
        },
        patterns=[],
        tree=False,
        output_json=False,
        output_ndjson=False,
        workspace_filter="~/code/",
        expected_repo_names=["flask"],
    ),
]


@pytest.mark.parametrize(
    list(ListReposFixture._fields),
    LIST_REPOS_FIXTURES,
    ids=[fixture.test_id for fixture in LIST_REPOS_FIXTURES],
)
def test_list_repos(
    test_id: str,
    config_data: dict[str, t.Any],
    patterns: list[str],
    tree: bool,
    output_json: bool,
    output_ndjson: bool,
    workspace_filter: str | None,
    expected_repo_names: list[str],
    tmp_path: pathlib.Path,
    monkeypatch: MonkeyPatch,
    capsys: t.Any,
) -> None:
    """Test listing repositories."""
    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.chdir(tmp_path)

    config_file = tmp_path / ".vcspull.yaml"
    create_test_config(config_file, config_data)

    # Run list_repos
    list_repos(
        repo_patterns=patterns,
        config_path=config_file,
        workspace_root=workspace_filter,
        tree=tree,
        output_json=output_json,
        output_ndjson=output_ndjson,
        color="never",
    )

    captured = capsys.readouterr()

    if output_json:
        # Parse JSON output
        output_data = json.loads(captured.out)
        assert isinstance(output_data, list)
        repo_names_in_output = [item["name"] for item in output_data]
        for expected_name in expected_repo_names:
            assert expected_name in repo_names_in_output
    elif output_ndjson:
        # Parse NDJSON output
        lines = [line for line in captured.out.strip().split("\n") if line]
        repo_names_in_output = [json.loads(line)["name"] for line in lines]
        for expected_name in expected_repo_names:
            assert expected_name in repo_names_in_output
    else:
        # Human-readable output
        for expected_name in expected_repo_names:
            assert expected_name in captured.out


def test_list_repos_tree_mode(
    tmp_path: pathlib.Path,
    monkeypatch: MonkeyPatch,
    capsys: t.Any,
) -> None:
    """Test listing repositories in tree mode."""
    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.chdir(tmp_path)

    config_file = tmp_path / ".vcspull.yaml"
    config_data = {
        "~/code/": {
            "flask": {"repo": "git+https://github.com/pallets/flask.git"},
        },
        "~/work/": {
            "myproject": {"repo": "git+https://github.com/user/myproject.git"},
        },
    }
    create_test_config(config_file, config_data)

    list_repos(
        repo_patterns=[],
        config_path=config_file,
        workspace_root=None,
        tree=True,
        output_json=False,
        output_ndjson=False,
        color="never",
    )

    captured = capsys.readouterr()

    # Should show repos (workspace roots may be normalized to paths)
    assert "flask" in captured.out
    assert "myproject" in captured.out
    # Tree mode should group repos
    assert "•" in captured.out


def test_list_repos_empty_config(
    tmp_path: pathlib.Path,
    monkeypatch: MonkeyPatch,
    capsys: t.Any,
) -> None:
    """Test listing with empty config shows appropriate message."""
    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.chdir(tmp_path)

    config_file = tmp_path / ".vcspull.yaml"
    create_test_config(config_file, {})

    list_repos(
        repo_patterns=[],
        config_path=config_file,
        workspace_root=None,
        tree=False,
        output_json=False,
        output_ndjson=False,
        color="never",
    )

    captured = capsys.readouterr()
    assert "No repositories found" in captured.out


def test_list_repos_pattern_no_match(
    tmp_path: pathlib.Path,
    monkeypatch: MonkeyPatch,
    capsys: t.Any,
) -> None:
    """Test listing with pattern that matches nothing."""
    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.chdir(tmp_path)

    config_file = tmp_path / ".vcspull.yaml"
    config_data = {
        "~/code/": {
            "flask": {"repo": "git+https://github.com/pallets/flask.git"},
        },
    }
    create_test_config(config_file, config_data)

    list_repos(
        repo_patterns=["nonexistent*"],
        config_path=config_file,
        workspace_root=None,
        tree=False,
        output_json=False,
        output_ndjson=False,
        color="never",
    )

    captured = capsys.readouterr()
    assert "No repositories found" in captured.out


# Tests for path contraction in JSON output


class PathContractionFixture(t.NamedTuple):
    """Fixture for testing path contraction in JSON/NDJSON output."""

    test_id: str
    output_json: bool
    output_ndjson: bool
    tree: bool


PATH_CONTRACTION_FIXTURES: list[PathContractionFixture] = [
    PathContractionFixture(
        test_id="json-output-contracts-paths",
        output_json=True,
        output_ndjson=False,
        tree=False,
    ),
    PathContractionFixture(
        test_id="ndjson-output-contracts-paths",
        output_json=False,
        output_ndjson=True,
        tree=False,
    ),
    PathContractionFixture(
        test_id="json-tree-output-contracts-paths",
        output_json=True,
        output_ndjson=False,
        tree=True,
    ),
]


@pytest.mark.parametrize(
    list(PathContractionFixture._fields),
    PATH_CONTRACTION_FIXTURES,
    ids=[fixture.test_id for fixture in PATH_CONTRACTION_FIXTURES],
)
def test_list_repos_path_contraction(
    test_id: str,
    output_json: bool,
    output_ndjson: bool,
    tree: bool,
    tmp_path: pathlib.Path,
    monkeypatch: MonkeyPatch,
    capsys: t.Any,
) -> None:
    """Test that JSON/NDJSON output contracts home directory paths."""
    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.chdir(tmp_path)

    config_file = tmp_path / ".vcspull.yaml"
    config_data = {
        "~/code/": {
            "flask": {"repo": "git+https://github.com/pallets/flask.git"},
            "django": {"repo": "git+https://github.com/django/django.git"},
        },
    }
    create_test_config(config_file, config_data)

    list_repos(
        repo_patterns=[],
        config_path=config_file,
        workspace_root=None,
        tree=tree,
        output_json=output_json,
        output_ndjson=output_ndjson,
        color="never",
    )

    captured = capsys.readouterr()

    if output_json:
        output_data = json.loads(captured.out)
        assert isinstance(output_data, list)
        for item in output_data:
            path = item["path"]
            # Path should start with ~/ not /home/<user>/
            assert path.startswith("~/"), f"Path {path} should be contracted to ~/..."
            assert not path.startswith(str(tmp_path)), (
                f"Path {path} should not contain absolute home path"
            )
    elif output_ndjson:
        lines = [line for line in captured.out.strip().split("\n") if line]
        for line in lines:
            item = json.loads(line)
            path = item["path"]
            # Path should start with ~/ not /home/<user>/
            assert path.startswith("~/"), f"Path {path} should be contracted to ~/..."
            assert not path.startswith(str(tmp_path)), (
                f"Path {path} should not contain absolute home path"
            )



"""Format vcspull configuration files."""

from __future__ import annotations

import argparse
import copy
import logging
import pathlib
import traceback
import typing as t

from colorama import Fore, Style

from vcspull._internal.config_reader import DuplicateAwareConfigReader
from vcspull._internal.private_path import PrivatePath
from vcspull.config import (
    find_config_files,
    find_home_config_files,
    merge_duplicate_workspace_roots,
    normalize_workspace_roots,
    save_config_yaml,
)

log = logging.getLogger(__name__)


def create_fmt_subparser(parser: argparse.ArgumentParser) -> None:
    """Create ``vcspull fmt`` argument subparser."""
    parser.add_argument(
        "-f",
        "--file",
        dest="config",
        metavar="FILE",
        help="path to config file (default: .vcspull.yaml or ~/.vcspull.yaml)",
    )
    parser.add_argument(
        "--write",
        "-w",
        action="store_true",
        help="Write formatted configuration back to file",
    )
    parser.add_argument(
        "--all",
        action="store_true",
        help="Format all discovered config files (home, config dir, and current dir)",
    )
    parser.add_argument(
        "--no-merge",
        dest="merge_roots",
        action="store_false",
        help="Do not merge duplicate workspace roots when formatting",
    )
    parser.set_defaults(merge_roots=True)


def normalize_repo_config(repo_data: t.Any) -> dict[str, t.Any]:
    """Normalize repository configuration to verbose format.

    Parameters
    ----------
    repo_data : Any
        Repository configuration (string URL or dict)

    Returns
    -------
    dict
        Normalized repository configuration with 'repo' key
    """
    if isinstance(repo_data, str):
        # Convert compact format to verbose format
        return {"repo": repo_data}
    if isinstance(repo_data, dict):
        # If it has 'url' key but not 'repo', convert to use 'repo'
        if "url" in repo_data and "repo" not in repo_data:
            normalized = repo_data.copy()
            normalized["repo"] = normalized.pop("url")
            return normalized
        # Already in correct format or has other fields
        return repo_data
    # Return as-is for other types
    return t.cast("dict[str, t.Any]", repo_data)


def format_config(config_data: dict[str, t.Any]) -> tuple[dict[str, t.Any], int]:
    """Format vcspull configuration for consistency.

    Parameters
    ----------
    config_data : dict
        Raw configuration data

    Returns
    -------
    tuple[dict, int]
        Formatted configuration and count of changes made
    """
    changes = 0
    formatted: dict[str, t.Any] = {}

    # Sort directories
    sorted_dirs = sorted(config_data.keys())

    for directory in sorted_dirs:
        repos = config_data[directory]

        if not isinstance(repos, dict):
            # Not a repository section, keep as-is
            formatted[directory] = repos
            continue

        # Sort repositories within each directory
        sorted_repos = sorted(repos.keys())
        formatted_dir: dict[str, t.Any] = {}

        for repo_name in sorted_repos:
            repo_data = repos[repo_name]
            normalized = normalize_repo_config(repo_data)

            # Check if normalization changed anything
            if normalized != repo_data:
                changes += 1

            formatted_dir[repo_name] = normalized

        # Check if sorting changed the order
        if list(repos.keys()) != sorted_repos:
            changes += 1

        formatted[directory] = formatted_dir

    # Check if directory sorting changed the order
    if list(config_data.keys()) != sorted_dirs:
        changes += 1

    return formatted, changes


def format_single_config(
    config_file_path: pathlib.Path,
    write: bool,
    *,
    merge_roots: bool,
) -> bool:
    """Format a single vcspull configuration file.

    Parameters
    ----------
    config_file_path : pathlib.Path
        Path to config file
    write : bool
        Whether to write changes back to file
    merge_roots : bool
        Merge duplicate workspace roots when True (default behavior)

    Returns
    -------
    bool
        True if formatting was successful, False otherwise
    """
    # Precompute redacted path for CLI output.
    display_config_path = str(PrivatePath(config_file_path))

    # Check if file exists
    if not config_file_path.exists():
        log.error(
            "%s✗%s Config file %s%s%s not found.",
            Fore.RED,
            Style.RESET_ALL,
            Fore.BLUE,
            display_config_path,
            Style.RESET_ALL,
        )
        return False

    # Load existing config
    try:
        raw_config, duplicate_root_occurrences, _top_level_items = (
            DuplicateAwareConfigReader.load_with_duplicates(config_file_path)
        )
    except TypeError:
        log.exception(
            "Config file %s is not a mapping",
            PrivatePath(config_file_path),
        )
        return False
    except Exception:
        log.exception(
            "Error loading config from %s",
            PrivatePath(config_file_path),
        )
        if log.isEnabledFor(logging.DEBUG):
            traceback.print_exc()
        return False

    # Format the configuration
    cwd = pathlib.Path.cwd()
    home = pathlib.Path.home()

    duplicate_merge_conflicts: list[str] = []
    duplicate_merge_changes = 0
    duplicate_merge_details: list[tuple[str, int]] = []

    working_config = copy.deepcopy(raw_config)

    if merge_roots:
        (
            working_config,
            duplicate_merge_conflicts,
            duplicate_merge_changes,
            duplicate_merge_details,
        ) = merge_duplicate_workspace_roots(working_config, duplicate_root_occurrences)
    elif duplicate_root_occurrences:
        duplicate_merge_details = [
            (label, len(values)) for label, values in duplicate_root_occurrences.items()
        ]
        for label, occurrence_count in duplicate_merge_details:
            log.warning(
                "%s•%s Duplicate workspace root %s%s%s appears %s%d%s time%s; "
                "skipping merge because --no-merge was provided.",
                Fore.BLUE,
                Style.RESET_ALL,
                Fore.MAGENTA,
                label,
                Style.RESET_ALL,
                Fore.YELLOW,
                occurrence_count,
                Style.RESET_ALL,
                "" if occurrence_count == 1 else "s",
            )

    if merge_roots:
        normalization_result = normalize_workspace_roots(
            working_config,
            cwd=cwd,
            home=home,
        )
        (
            normalized_config,
            _workspace_map,
            merge_conflicts,
            normalization_changes,
        ) = normalization_result
    else:
        normalized_config = working_config
        merge_conflicts = []
        normalization_changes = 0

    for message in merge_conflicts:
        log.warning(message)
    for message in duplicate_merge_conflicts:
        log.warning(message)

    formatted_config, change_count = format_config(normalized_config)
    change_count += normalization_changes + duplicate_merge_changes

    if change_count == 0:
        log.info(
            "%s✓%s %s%s%s is already formatted correctly.",
            Fore.GREEN,
            Style.RESET_ALL,
            Fore.BLUE,
            display_config_path,
            Style.RESET_ALL,
        )
        return True

    # Show what would be changed
    log.info(
        "%si%s Found %s%d%s formatting %s in %s%s%s",
        Fore.CYAN,
        Style.RESET_ALL,
        Fore.YELLOW,
        change_count,
        Style.RESET_ALL,
        "issue" if change_count == 1 else "issues",
        Fore.BLUE,
        display_config_path,
        Style.RESET_ALL,
    )

    # Analyze and report specific changes
    if merge_roots and normalization_changes > 0:
        log.info(
            "  %s•%s Normalized workspace root labels",
            Fore.BLUE,
            Style.RESET_ALL,
        )

    if merge_roots and duplicate_merge_details:
        for label, occurrence_count in duplicate_merge_details:
            log.info(
                "  %s•%s Merged %s%d%s duplicate entr%s for workspace root %s%s%s",
                Fore.BLUE,
                Style.RESET_ALL,
                Fore.YELLOW,
                occurrence_count,
                Style.RESET_ALL,
                "y" if occurrence_count == 1 else "ies",
                Fore.MAGENTA,
                label,
                Style.RESET_ALL,
            )

    compact_to_verbose = 0
    url_to_repo = 0

    for repos in normalized_config.values():
        if isinstance(repos, dict):
            for repo_data in repos.values():
                if isinstance(repo_data, str):
                    compact_to_verbose += 1
                elif (
                    isinstance(repo_data, dict)
                    and "url" in repo_data
                    and "repo" not in repo_data
                ):
                    url_to_repo += 1

    if compact_to_verbose > 0:
        log.info(
            "  %s•%s %d %s from compact to verbose format",
            Fore.BLUE,
            Style.RESET_ALL,
            compact_to_verbose,
            "repository" if compact_to_verbose == 1 else "repositories",
        )

    if url_to_repo > 0:
        log.info(
            "  %s•%s %d %s from 'url' to 'repo' key",
            Fore.BLUE,
            Style.RESET_ALL,
            url_to_repo,
            "repository" if url_to_repo == 1 else "repositories",
        )

    if list(normalized_config.keys()) != sorted(normalized_config.keys()):
        log.info(
            "  %s•%s Directories will be sorted alphabetically",
            Fore.BLUE,
            Style.RESET_ALL,
        )

    # Check if any repos need sorting
    for directory, repos in normalized_config.items():
        if isinstance(repos, dict) and list(repos.keys()) != sorted(repos.keys()):
            log.info(
                "  %s•%s Repositories in %s%s%s will be sorted alphabetically",
                Fore.BLUE,
                Style.RESET_ALL,
                Fore.MAGENTA,
                directory,
                Style.RESET_ALL,
            )
            break

    if write:
        # Save formatted config
        try:
            save_config_yaml(config_file_path, formatted_config)
            log.info(
                "%s✓%s Successfully formatted %s%s%s",
                Fore.GREEN,
                Style.RESET_ALL,
                Fore.BLUE,
                display_config_path,
                Style.RESET_ALL,
            )
        except Exception:
            log.exception(
                "Error saving formatted config to %s",
                PrivatePath(config_file_path),
            )
            if log.isEnabledFor(logging.DEBUG):
                traceback.print_exc()
            return False
    else:
        log.info(
            "\n%s→%s Run with %s--write%s to apply these formatting changes.",
            Fore.YELLOW,
            Style.RESET_ALL,
            Fore.CYAN,
            Style.RESET_ALL,
        )

    return True


def format_config_file(
    config_file_path_str: str | None,
    write: bool,
    format_all: bool = False,
    *,
    merge_roots: bool = True,
) -> None:
    """Format vcspull configuration file(s).

    Parameters
    ----------
    config_file_path_str : str | None
        Path to config file, or None to use default
    write : bool
        Whether to write changes back to file
    format_all : bool
        If True, format all discovered config files
    merge_roots : bool
        Merge duplicate workspace roots when True (default)
    """
    if format_all:
        # Format all discovered config files
        config_files = find_config_files(include_home=True)

        # Also check for local .vcspull.yaml
        local_yaml = pathlib.Path.cwd() / ".vcspull.yaml"
        if local_yaml.exists() and local_yaml not in config_files:
            config_files.append(local_yaml)

        # Also check for local .vcspull.json
        local_json = pathlib.Path.cwd() / ".vcspull.json"
        if local_json.exists() and local_json not in config_files:
            config_files.append(local_json)

        if not config_files:
            log.error(
                "%s✗%s No configuration files found.",
                Fore.RED,
                Style.RESET_ALL,
            )
            return

        log.info(
            "%si%s Found %s%d%s configuration %s to format:",
            Fore.CYAN,
            Style.RESET_ALL,
            Fore.YELLOW,
            len(config_files),
            Style.RESET_ALL,
            "file" if len(config_files) == 1 else "files",
        )

        for config_file in config_files:
            display_config_file = str(PrivatePath(config_file))
            log.info(
                "  %s•%s %s%s%s",
                Fore.BLUE,
                Style.RESET_ALL,
                Fore.CYAN,
                display_config_file,
                Style.RESET_ALL,
            )

        log.info("")  # Empty line for readability

        success_count = 0
        for config_file in config_files:
            if format_single_config(
                config_file,
                write,
                merge_roots=merge_roots,
            ):
                success_count += 1

        # Summary
        if success_count == len(config_files):
            log.info(
                "\n%s✓%s All %d configuration files processed successfully.",
                Fore.GREEN,
                Style.RESET_ALL,
                len(config_files),
            )
        else:
            log.info(
                "\n%si%s Processed %d/%d configuration files successfully.",
                Fore.CYAN,
                Style.RESET_ALL,
                success_count,
                len(config_files),
            )
    else:
        # Format single config file
        if config_file_path_str:
            config_file_path = pathlib.Path(config_file_path_str).expanduser().resolve()
        else:
            home_configs = find_home_config_files(filetype=["yaml"])
            if not home_configs:
                # Try local .vcspull.yaml
                local_config = pathlib.Path.cwd() / ".vcspull.yaml"
                if local_config.exists():
                    config_file_path = local_config
                else:
                    log.error(
                        "%s✗%s No configuration file found. "
                        "Create .vcspull.yaml first.",
                        Fore.RED,
                        Style.RESET_ALL,
                    )
                    return
            elif len(home_configs) > 1:
                log.error(
                    "Multiple home config files found, "
                    "please specify one with -f/--file",
                )
                return
            else:
                config_file_path = home_configs[0]

        format_single_config(
            config_file_path,
            write,
            merge_roots=merge_roots,
        )



"""Add single repository functionality for vcspull."""

from __future__ import annotations

import argparse
import copy
import logging
import pathlib
import subprocess
import traceback
import typing as t

from colorama import Fore, Style

from vcspull._internal.config_reader import DuplicateAwareConfigReader
from vcspull._internal.private_path import PrivatePath
from vcspull.config import (
    canonicalize_workspace_path,
    expand_dir,
    find_home_config_files,
    merge_duplicate_workspace_roots,
    save_config_yaml,
    save_config_yaml_with_items,
    workspace_root_label,
)

log = logging.getLogger(__name__)


def create_add_subparser(parser: argparse.ArgumentParser) -> None:
    """Create ``vcspull add`` argument subparser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser to configure
    """
    parser.add_argument(
        "repo_path",
        help=(
            "Filesystem path to an existing project. The parent directory "
            "becomes the workspace unless overridden with --workspace."
        ),
    )
    parser.add_argument(
        "--name",
        dest="override_name",
        help="Override detected repository name when importing from a path",
    )
    parser.add_argument(
        "--url",
        dest="url",
        help="Repository URL to record (overrides detected remotes)",
    )
    parser.add_argument(
        "-f",
        "--file",
        dest="config",
        metavar="FILE",
        help="path to config file (default: ~/.vcspull.yaml or ./.vcspull.yaml)",
    )
    parser.add_argument(
        "-w",
        "--workspace",
        "--workspace-root",
        dest="workspace_root_path",
        metavar="DIR",
        help=(
            "Workspace root directory in config (e.g., '~/projects/'). Defaults "
            "to the parent directory of the repository path."
        ),
    )
    parser.add_argument(
        "--dry-run",
        "-n",
        action="store_true",
        help="Preview changes without writing to config file",
    )
    parser.add_argument(
        "--no-merge",
        dest="merge_duplicates",
        action="store_false",
        help="Skip merging duplicate workspace roots before writing",
    )
    parser.add_argument(
        "-y",
        "--yes",
        dest="assume_yes",
        action="store_true",
        help="Automatically confirm interactive prompts",
    )
    parser.set_defaults(merge_duplicates=True)


def _resolve_workspace_path(
    workspace_root: str | None,
    repo_path_str: str | None,
    *,
    cwd: pathlib.Path,
) -> pathlib.Path:
    """Resolve workspace path from arguments.

    Parameters
    ----------
    workspace_root : str | None
        Workspace root path from user
    repo_path_str : str | None
        Repo path from user
    cwd : pathlib.Path
        Current working directory

    Returns
    -------
    pathlib.Path
        Resolved workspace path
    """
    if workspace_root:
        return canonicalize_workspace_path(workspace_root, cwd=cwd)
    if repo_path_str:
        repo_path = expand_dir(pathlib.Path(repo_path_str), cwd)
        return repo_path.parent
    return cwd


def _detect_git_remote(repo_path: pathlib.Path) -> str | None:
    """Return the ``origin`` remote URL for a Git repository if available."""
    try:
        result = subprocess.run(
            ["git", "-C", str(repo_path), "remote", "get-url", "origin"],
            check=True,
            capture_output=True,
            text=True,
        )
    except FileNotFoundError:
        log.debug("git executable not found when inspecting %s", repo_path)
        return None
    except subprocess.CalledProcessError:
        log.debug("No git remote 'origin' configured for %s", repo_path)
        return None

    remote = result.stdout.strip()
    return remote or None


def _normalize_detected_url(remote: str | None) -> tuple[str, str]:
    """Return display and config URLs derived from a detected remote."""
    if remote is None:
        return "", ""

    display_url = remote
    config_url = remote

    normalized = remote.strip()

    if normalized and not normalized.startswith("git+"):
        if normalized.startswith(("http://", "https://", "file://")):
            config_url = f"git+{normalized}"
        else:
            config_url = normalized
    elif normalized:
        config_url = normalized

    return display_url, config_url


def _build_ordered_items(
    top_level_items: list[tuple[str, t.Any]] | None,
    raw_config: dict[str, t.Any],
) -> list[dict[str, t.Any]]:
    """Return deep-copied top-level items preserving original ordering."""
    source: list[tuple[str, t.Any]] = top_level_items or list(raw_config.items())

    ordered: list[dict[str, t.Any]] = []
    for label, section in source:
        ordered.append({"label": label, "section": copy.deepcopy(section)})
    return ordered


def _aggregate_from_ordered_items(
    items: list[dict[str, t.Any]],
) -> dict[str, t.Any]:
    """Collapse ordered top-level items into a mapping grouped by label."""
    aggregated: dict[str, t.Any] = {}
    for entry in items:
        label = entry["label"]
        section = entry["section"]
        if isinstance(section, dict):
            workspace_section = aggregated.setdefault(label, {})
            for repo_name, repo_config in section.items():
                workspace_section[repo_name] = copy.deepcopy(repo_config)
        else:
            aggregated[label] = copy.deepcopy(section)
    return aggregated


def _collect_duplicate_sections(
    items: list[dict[str, t.Any]],
) -> dict[str, list[t.Any]]:
    """Return mapping of labels to their repeated sections (>= 2 occurrences)."""
    occurrences: dict[str, list[t.Any]] = {}
    for entry in items:
        label = entry["label"]
        occurrences.setdefault(label, []).append(copy.deepcopy(entry["section"]))

    return {
        label: sections for label, sections in occurrences.items() if len(sections) > 1
    }


def handle_add_command(args: argparse.Namespace) -> None:
    """Entry point for the ``vcspull add`` CLI command."""
    repo_input = getattr(args, "repo_path", None)
    if repo_input is None:
        log.error("A repository path must be provided.")
        return

    cwd = pathlib.Path.cwd()
    repo_path = expand_dir(pathlib.Path(repo_input), cwd=cwd)

    if not repo_path.exists():
        log.error("Repository path %s does not exist.", PrivatePath(repo_path))
        return

    if not repo_path.is_dir():
        log.error("Repository path %s is not a directory.", PrivatePath(repo_path))
        return

    override_name = getattr(args, "override_name", None)
    repo_name = override_name or repo_path.name

    explicit_url = getattr(args, "url", None)
    if explicit_url:
        display_url, config_url = _normalize_detected_url(explicit_url)
    else:
        detected_remote = _detect_git_remote(repo_path)
        display_url, config_url = _normalize_detected_url(detected_remote)

    if not config_url:
        display_url = str(PrivatePath(repo_path))
        config_url = str(repo_path)
        log.warning(
            "Unable to determine git remote for %s; using local path in config.",
            repo_path,
        )

    workspace_root_arg = getattr(args, "workspace_root_path", None)
    workspace_root_input = (
        workspace_root_arg
        if workspace_root_arg is not None
        else repo_path.parent.as_posix()
    )

    workspace_path = expand_dir(pathlib.Path(workspace_root_input), cwd=cwd)
    workspace_label = workspace_root_label(
        workspace_path,
        cwd=cwd,
        home=pathlib.Path.home(),
        preserve_cwd_label=workspace_root_arg in {".", "./"},
    )

    summary_url = display_url or config_url

    display_path = str(PrivatePath(repo_path))

    log.info("%sFound new repository to import:%s", Fore.GREEN, Style.RESET_ALL)
    log.info(
        "  %s+%s %s%s%s (%s%s%s)",
        Fore.GREEN,
        Style.RESET_ALL,
        Fore.CYAN,
        repo_name,
        Style.RESET_ALL,
        Fore.YELLOW,
        summary_url,
        Style.RESET_ALL,
    )
    log.info(
        "  %s•%s workspace: %s%s%s",
        Fore.BLUE,
        Style.RESET_ALL,
        Fore.MAGENTA,
        workspace_label,
        Style.RESET_ALL,
    )
    log.info(
        "  %s↳%s path: %s%s%s",
        Fore.BLUE,
        Style.RESET_ALL,
        Fore.BLUE,
        display_path,
        Style.RESET_ALL,
    )

    prompt_text = f"{Fore.CYAN}?{Style.RESET_ALL} Import this repository? [y/N]: "

    proceed = True
    if args.dry_run:
        log.info(
            "%s?%s Import this repository? [y/N]: %sskipped (dry-run)%s",
            Fore.CYAN,
            Style.RESET_ALL,
            Fore.YELLOW,
            Style.RESET_ALL,
        )
    elif getattr(args, "assume_yes", False):
        log.info(
            "%s?%s Import this repository? [y/N]: %sy (auto-confirm)%s",
            Fore.CYAN,
            Style.RESET_ALL,
            Fore.GREEN,
            Style.RESET_ALL,
        )
    else:
        try:
            response = input(prompt_text)
        except EOFError:
            response = ""
        proceed = response.strip().lower() in {"y", "yes"}
        if not proceed:
            log.info(
                "Aborted import of '%s' from %s",
                repo_name,
                PrivatePath(repo_path),
            )
            return

    add_repo(
        name=repo_name,
        url=config_url,
        config_file_path_str=args.config,
        path=str(repo_path),
        workspace_root_path=workspace_root_input,
        dry_run=args.dry_run,
        merge_duplicates=args.merge_duplicates,
    )


def add_repo(
    name: str,
    url: str,
    config_file_path_str: str | None,
    path: str | None,
    workspace_root_path: str | None,
    dry_run: bool,
    *,
    merge_duplicates: bool = True,
) -> None:
    """Add a repository to the vcspull configuration.

    Parameters
    ----------
    name : str
        Repository name for the config
    url : str
        Repository URL
    config_file_path_str : str | None
        Path to config file, or None to use default
    path : str | None
        Local path where repo will be cloned
    workspace_root_path : str | None
        Workspace root to use in config
    dry_run : bool
        If True, preview changes without writing
    """
    # Determine config file
    config_file_path: pathlib.Path
    if config_file_path_str:
        config_file_path = pathlib.Path(config_file_path_str).expanduser().resolve()
    else:
        home_configs = find_home_config_files(filetype=["yaml"])
        if not home_configs:
            config_file_path = pathlib.Path.cwd() / ".vcspull.yaml"
            log.info(
                "No config specified and no default found, will create at %s",
                PrivatePath(config_file_path),
            )
        elif len(home_configs) > 1:
            log.error(
                "Multiple home config files found, please specify one with -f/--file",
            )
            return
        else:
            config_file_path = home_configs[0]

    # Load existing config
    raw_config: dict[str, t.Any]
    duplicate_root_occurrences: dict[str, list[t.Any]]
    top_level_items: list[tuple[str, t.Any]]
    display_config_path = str(PrivatePath(config_file_path))

    if config_file_path.exists() and config_file_path.is_file():
        try:
            (
                raw_config,
                duplicate_root_occurrences,
                top_level_items,
            ) = DuplicateAwareConfigReader.load_with_duplicates(config_file_path)
        except TypeError:
            log.exception(
                "Config file %s is not a valid YAML dictionary.",
                display_config_path,
            )
            return
        except Exception:
            log.exception(
                "Error loading YAML from %s. Aborting.",
                PrivatePath(config_file_path),
            )
            if log.isEnabledFor(logging.DEBUG):
                traceback.print_exc()
            return
    else:
        raw_config = {}
        duplicate_root_occurrences = {}
        top_level_items = []
        log.info(
            "Config file %s not found. A new one will be created.",
            display_config_path,
        )

    cwd = pathlib.Path.cwd()
    home = pathlib.Path.home()

    workspace_path = _resolve_workspace_path(
        workspace_root_path,
        path,
        cwd=cwd,
    )

    explicit_dot = workspace_root_path in {".", "./"}

    preferred_label = workspace_root_label(
        workspace_path,
        cwd=cwd,
        home=home,
        preserve_cwd_label=explicit_dot,
    )

    new_repo_entry = {"repo": url}

    def _ensure_workspace_label_for_merge(
        config_data: dict[str, t.Any],
    ) -> tuple[str, bool]:
        workspace_map: dict[pathlib.Path, str] = {}
        for label, section in config_data.items():
            if not isinstance(section, dict):
                continue
            try:
                path_key = canonicalize_workspace_path(label, cwd=cwd)
            except Exception:
                continue
            workspace_map[path_key] = label

        existing_label = workspace_map.get(workspace_path)
        relabelled = False

        if explicit_dot:
            workspace_label = "./"
            if existing_label and existing_label != "./":
                config_data["./"] = config_data.pop(existing_label)
                relabelled = True
            else:
                config_data.setdefault("./", {})
        elif existing_label is None:
            workspace_label = preferred_label
            config_data.setdefault(workspace_label, {})
        else:
            workspace_label = existing_label

        if workspace_label not in config_data:
            config_data[workspace_label] = {}

        return workspace_label, relabelled

    def _prepare_no_merge_items(
        items: list[dict[str, t.Any]],
    ) -> tuple[str, int, bool]:
        matching_indexes: list[int] = []
        for idx, entry in enumerate(items):
            section = entry["section"]
            if not isinstance(section, dict):
                continue
            try:
                path_key = canonicalize_workspace_path(entry["label"], cwd=cwd)
            except Exception:
                continue
            if path_key == workspace_path:
                matching_indexes.append(idx)

        relabelled = False

        if explicit_dot:
            if matching_indexes:
                for idx in matching_indexes:
                    if items[idx]["label"] != "./":
                        items[idx]["label"] = "./"
                        relabelled = True
                target_index = matching_indexes[-1]
            else:
                items.append({"label": "./", "section": {}})
                target_index = len(items) - 1
            workspace_label = items[target_index]["label"]
            return workspace_label, target_index, relabelled

        if not matching_indexes:
            workspace_label = preferred_label
            items.append({"label": workspace_label, "section": {}})
            target_index = len(items) - 1
            return workspace_label, target_index, relabelled

        target_index = matching_indexes[-1]
        workspace_label = items[target_index]["label"]
        return workspace_label, target_index, relabelled

    config_was_relabelled = False
    duplicate_merge_conflicts: list[str] = []
    duplicate_merge_changes = 0
    duplicate_merge_details: list[tuple[str, int]] = []

    if merge_duplicates:
        (
            raw_config,
            duplicate_merge_conflicts,
            duplicate_merge_changes,
            duplicate_merge_details,
        ) = merge_duplicate_workspace_roots(raw_config, duplicate_root_occurrences)
        for message in duplicate_merge_conflicts:
            log.warning(message)

        if duplicate_merge_changes and duplicate_merge_details:
            for label, occurrence_count in duplicate_merge_details:
                log.info(
                    "%s•%s Merged %s%d%s duplicate entr%s for workspace root %s%s%s",
                    Fore.BLUE,
                    Style.RESET_ALL,
                    Fore.YELLOW,
                    occurrence_count,
                    Style.RESET_ALL,
                    "y" if occurrence_count == 1 else "ies",
                    Fore.MAGENTA,
                    label,
                    Style.RESET_ALL,
                )

        workspace_label, relabelled = _ensure_workspace_label_for_merge(raw_config)
        config_was_relabelled = relabelled
        workspace_section = raw_config.get(workspace_label)
        if not isinstance(workspace_section, dict):
            log.error(
                "Workspace root '%s' in configuration is not a dictionary. Aborting.",
                workspace_label,
            )
            return

        existing_config = workspace_section.get(name)
        if existing_config is not None:
            if isinstance(existing_config, str):
                current_url = existing_config
            elif isinstance(existing_config, dict):
                repo_value = existing_config.get("repo")
                url_value = existing_config.get("url")
                current_url = repo_value or url_value or "unknown"
            else:
                current_url = str(existing_config)

            log.warning(
                "Repository '%s' already exists under '%s'. Current URL: %s. "
                "To update, remove and re-add, or edit the YAML file manually.",
                name,
                workspace_label,
                current_url,
            )

            if (duplicate_merge_changes > 0 or config_was_relabelled) and not dry_run:
                try:
                    save_config_yaml(config_file_path, raw_config)
                    log.info(
                        "%s✓%s Workspace label adjustments saved to %s%s%s.",
                        Fore.GREEN,
                        Style.RESET_ALL,
                        Fore.BLUE,
                        display_config_path,
                        Style.RESET_ALL,
                    )
                except Exception:
                    log.exception(
                        "Error saving config to %s",
                        PrivatePath(config_file_path),
                    )
                    if log.isEnabledFor(logging.DEBUG):
                        traceback.print_exc()
            elif (duplicate_merge_changes > 0 or config_was_relabelled) and dry_run:
                log.info(
                    "%s→%s Would save workspace label adjustments to %s%s%s.",
                    Fore.YELLOW,
                    Style.RESET_ALL,
                    Fore.BLUE,
                    display_config_path,
                    Style.RESET_ALL,
                )
            return

        workspace_section[name] = copy.deepcopy(new_repo_entry)

        if dry_run:
            log.info(
                "%s→%s Would add %s'%s'%s (%s%s%s) to %s%s%s under '%s%s%s'.",
                Fore.YELLOW,
                Style.RESET_ALL,
                Fore.CYAN,
                name,
                Style.RESET_ALL,
                Fore.YELLOW,
                url,
                Style.RESET_ALL,
                Fore.BLUE,
                display_config_path,
                Style.RESET_ALL,
                Fore.MAGENTA,
                workspace_label,
                Style.RESET_ALL,
            )
            return

        try:
            save_config_yaml(config_file_path, raw_config)
            log.info(
                "%s✓%s Successfully added %s'%s'%s (%s%s%s) to %s%s%s under '%s%s%s'.",
                Fore.GREEN,
                Style.RESET_ALL,
                Fore.CYAN,
                name,
                Style.RESET_ALL,
                Fore.YELLOW,
                url,
                Style.RESET_ALL,
                Fore.BLUE,
                display_config_path,
                Style.RESET_ALL,
                Fore.MAGENTA,
                workspace_label,
                Style.RESET_ALL,
            )
        except Exception:
            log.exception(
                "Error saving config to %s",
                PrivatePath(config_file_path),
            )
            if log.isEnabledFor(logging.DEBUG):
                traceback.print_exc()
        return

    ordered_items = _build_ordered_items(top_level_items, raw_config)

    workspace_label, target_index, relabelled = _prepare_no_merge_items(ordered_items)
    config_was_relabelled = relabelled

    duplicate_sections = _collect_duplicate_sections(ordered_items)
    for label, sections in duplicate_sections.items():
        occurrence_count = len(sections)
        log.warning(
            "%s•%s Duplicate workspace root %s%s%s appears %s%d%s time%s; "
            "skipping merge because --no-merge was provided.",
            Fore.BLUE,
            Style.RESET_ALL,
            Fore.MAGENTA,
            label,
            Style.RESET_ALL,
            Fore.YELLOW,
            occurrence_count,
            Style.RESET_ALL,
            "" if occurrence_count == 1 else "s",
        )

    raw_config_view = _aggregate_from_ordered_items(ordered_items)
    workspace_section_view = raw_config_view.get(workspace_label)
    if workspace_section_view is None:
        workspace_section_view = {}
        raw_config_view[workspace_label] = workspace_section_view

    if not isinstance(workspace_section_view, dict):
        log.error(
            "Workspace root '%s' in configuration is not a dictionary. Aborting.",
            workspace_label,
        )
        return

    existing_config = workspace_section_view.get(name)
    if existing_config is not None:
        if isinstance(existing_config, str):
            current_url = existing_config
        elif isinstance(existing_config, dict):
            repo_value = existing_config.get("repo")
            url_value = existing_config.get("url")
            current_url = repo_value or url_value or "unknown"
        else:
            current_url = str(existing_config)

        log.warning(
            "Repository '%s' already exists under '%s'. Current URL: %s. "
            "To update, remove and re-add, or edit the YAML file manually.",
            name,
            workspace_label,
            current_url,
        )

        if config_was_relabelled:
            if dry_run:
                log.info(
                    "%s→%s Would save workspace label adjustments to %s%s%s.",
                    Fore.YELLOW,
                    Style.RESET_ALL,
                    Fore.BLUE,
                    display_config_path,
                    Style.RESET_ALL,
                )
            else:
                try:
                    save_config_yaml_with_items(
                        config_file_path,
                        [(entry["label"], entry["section"]) for entry in ordered_items],
                    )
                    log.info(
                        "%s✓%s Workspace label adjustments saved to %s%s%s.",
                        Fore.GREEN,
                        Style.RESET_ALL,
                        Fore.BLUE,
                        display_config_path,
                        Style.RESET_ALL,
                    )
                except Exception:
                    log.exception(
                        "Error saving config to %s",
                        PrivatePath(config_file_path),
                    )
                    if log.isEnabledFor(logging.DEBUG):
                        traceback.print_exc()
        return

    target_section = ordered_items[target_index]["section"]
    if not isinstance(target_section, dict):
        log.error(
            "Workspace root '%s' in configuration is not a dictionary. Aborting.",
            ordered_items[target_index]["label"],
        )
        return

    target_section[name] = copy.deepcopy(new_repo_entry)
    workspace_section_view[name] = copy.deepcopy(new_repo_entry)

    if dry_run:
        log.info(
            "%s→%s Would add %s'%s'%s (%s%s%s) to %s%s%s under '%s%s%s'.",
            Fore.YELLOW,
            Style.RESET_ALL,
            Fore.CYAN,
            name,
            Style.RESET_ALL,
            Fore.YELLOW,
            url,
            Style.RESET_ALL,
            Fore.BLUE,
            display_config_path,
            Style.RESET_ALL,
            Fore.MAGENTA,
            workspace_label,
            Style.RESET_ALL,
        )
        return

    try:
        save_config_yaml_with_items(
            config_file_path,
            [(entry["label"], entry["section"]) for entry in ordered_items],
        )
        log.info(
            "%s✓%s Successfully added %s'%s'%s (%s%s%s) to %s%s%s under '%s%s%s'.",
            Fore.GREEN,
            Style.RESET_ALL,
            Fore.CYAN,
            name,
            Style.RESET_ALL,
            Fore.YELLOW,
            url,
            Style.RESET_ALL,
            Fore.BLUE,
            display_config_path,
            Style.RESET_ALL,
            Fore.MAGENTA,
            workspace_label,
            Style.RESET_ALL,
        )
    except Exception:
        log.exception(
            "Error saving config to %s",
            PrivatePath(config_file_path),
        )
        if log.isEnabledFor(logging.DEBUG):
            traceback.print_exc()



"""Discover repositories from filesystem for vcspull."""

from __future__ import annotations

import argparse
import logging
import os
import pathlib
import subprocess
import traceback
import typing as t

from colorama import Fore, Style

from vcspull._internal.config_reader import DuplicateAwareConfigReader
from vcspull._internal.private_path import PrivatePath
from vcspull.config import (
    canonicalize_workspace_path,
    expand_dir,
    find_home_config_files,
    merge_duplicate_workspace_roots,
    normalize_workspace_roots,
    save_config_yaml,
    workspace_root_label,
)

log = logging.getLogger(__name__)

ConfigScope = t.Literal["system", "user", "project", "external"]


def _classify_config_scope(
    config_path: pathlib.Path,
    *,
    cwd: pathlib.Path,
    home: pathlib.Path,
) -> ConfigScope:
    """Determine whether a config lives in user, system, project, or external scope."""
    resolved = config_path.expanduser().resolve()
    home = home.expanduser().resolve()
    cwd = cwd.expanduser().resolve()

    default_user_configs = {
        (home / ".vcspull.yaml").resolve(),
        (home / ".vcspull.json").resolve(),
    }
    if resolved in default_user_configs:
        return "user"

    xdg_config_home = (
        pathlib.Path(os.environ.get("XDG_CONFIG_HOME", home / ".config"))
        .expanduser()
        .resolve()
    )
    user_config_root = (xdg_config_home / "vcspull").resolve()
    try:
        resolved.relative_to(user_config_root)
    except ValueError:
        pass
    else:
        return "user"

    xdg_config_dirs_value = os.environ.get("XDG_CONFIG_DIRS")
    if xdg_config_dirs_value:
        config_dir_bases = [
            pathlib.Path(entry).expanduser().resolve()
            for entry in xdg_config_dirs_value.split(os.pathsep)
            if entry
        ]
    else:
        config_dir_bases = [pathlib.Path("/etc/xdg").resolve()]

    for base in config_dir_bases:
        candidate = (base / "vcspull").resolve()
        try:
            resolved.relative_to(candidate)
        except ValueError:
            continue
        else:
            return "system"

    try:
        resolved.relative_to(cwd)
    except ValueError:
        return "external"
    return "project"


def get_git_origin_url(repo_path: pathlib.Path) -> str | None:
    """Get the origin URL from a git repository.

    Parameters
    ----------
    repo_path : pathlib.Path
        Path to the git repository

    Returns
    -------
    str | None
        The origin URL if found, None otherwise
    """
    try:
        result = subprocess.run(
            ["git", "config", "--get", "remote.origin.url"],
            cwd=repo_path,
            capture_output=True,
            text=True,
            check=True,
        )
        return result.stdout.strip()
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        log.debug("Could not get origin URL for %s: %s", repo_path, e)
        return None


def create_discover_subparser(parser: argparse.ArgumentParser) -> None:
    """Create ``vcspull discover`` argument subparser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser to configure
    """
    parser.add_argument(
        "scan_dir",
        metavar="PATH",
        help="Directory to scan for git repositories",
    )
    parser.add_argument(
        "-f",
        "--file",
        dest="config",
        metavar="FILE",
        help="path to config file (default: ~/.vcspull.yaml or ./.vcspull.yaml)",
    )
    parser.add_argument(
        "-w",
        "--workspace",
        "--workspace-root",
        dest="workspace_root_path",
        metavar="DIR",
        help=(
            "Workspace root directory in config (e.g., '~/projects/'). "
            "If not specified, uses the scan directory. "
            "Applies the workspace root to all discovered repos."
        ),
    )
    parser.add_argument(
        "--recursive",
        "-r",
        action="store_true",
        help="Scan directories recursively",
    )
    parser.add_argument(
        "--yes",
        "-y",
        action="store_true",
        help="Skip confirmation prompt",
    )
    parser.add_argument(
        "--dry-run",
        "-n",
        action="store_true",
        help="Preview changes without writing to config file",
    )
    parser.add_argument(
        "--no-merge",
        dest="merge_duplicates",
        action="store_false",
        help="Skip merging duplicate workspace roots before writing",
    )
    parser.set_defaults(merge_duplicates=True)


def _resolve_workspace_path(
    workspace_root: str | None,
    repo_path_str: str | None,
    *,
    cwd: pathlib.Path,
) -> pathlib.Path:
    """Resolve workspace path from arguments.

    Parameters
    ----------
    workspace_root : str | None
        Workspace root path from user
    repo_path_str : str | None
        Repo path from user
    cwd : pathlib.Path
        Current working directory

    Returns
    -------
    pathlib.Path
        Resolved workspace path
    """
    if workspace_root:
        return canonicalize_workspace_path(workspace_root, cwd=cwd)
    if repo_path_str:
        return expand_dir(pathlib.Path(repo_path_str), cwd)
    return cwd


def discover_repos(
    scan_dir_str: str,
    config_file_path_str: str | None,
    recursive: bool,
    workspace_root_override: str | None,
    yes: bool,
    dry_run: bool,
    *,
    merge_duplicates: bool = True,
) -> None:
    """Scan filesystem for git repositories and add to vcspull config.

    Parameters
    ----------
    scan_dir_str : str
        Directory to scan for git repositories
    config_file_path_str : str | None
        Path to config file, or None to use default
    recursive : bool
        Whether to scan subdirectories recursively
    workspace_root_override : str | None
        Workspace root to use in config (overrides automatic detection)
    yes : bool
        Whether to skip confirmation prompt
    dry_run : bool
        If True, preview changes without writing
    """
    scan_dir = expand_dir(pathlib.Path(scan_dir_str))

    config_file_path: pathlib.Path
    if config_file_path_str:
        config_file_path = pathlib.Path(config_file_path_str).expanduser().resolve()
    else:
        home_configs = find_home_config_files(filetype=["yaml"])
        if not home_configs:
            config_file_path = pathlib.Path.cwd() / ".vcspull.yaml"
            log.info(
                "%si%s No config specified and no default "
                "home config, will use/create %s%s%s",
                Fore.CYAN,
                Style.RESET_ALL,
                Fore.BLUE,
                PrivatePath(config_file_path),
                Style.RESET_ALL,
            )
        elif len(home_configs) > 1:
            log.error(
                "Multiple home_config files found, please specify one with -f/--file",
            )
            return
        else:
            config_file_path = home_configs[0]

    display_config_path = str(PrivatePath(config_file_path))

    cwd = pathlib.Path.cwd()
    home = pathlib.Path.home()
    config_scope = _classify_config_scope(config_file_path, cwd=cwd, home=home)
    allow_relative_workspace = config_scope == "project"

    raw_config: dict[str, t.Any]
    duplicate_root_occurrences: dict[str, list[t.Any]]
    if config_file_path.exists() and config_file_path.is_file():
        try:
            (
                raw_config,
                duplicate_root_occurrences,
                _top_level_items,
            ) = DuplicateAwareConfigReader.load_with_duplicates(config_file_path)
        except TypeError:
            log.exception(
                "Config file %s is not a valid YAML dictionary.",
                display_config_path,
            )
            return
        except Exception:
            log.exception(
                "Error loading YAML from %s. Aborting.",
                PrivatePath(config_file_path),
            )
            if log.isEnabledFor(logging.DEBUG):
                traceback.print_exc()
            return
        if raw_config is None:
            raw_config = {}
        elif not isinstance(raw_config, dict):
            log.error(
                "Config file %s is not a valid YAML dictionary.",
                display_config_path,
            )
            return
    else:
        raw_config = {}
        duplicate_root_occurrences = {}
        log.info(
            "%si%s Config file %s%s%s not found. A new one will be created.",
            Fore.CYAN,
            Style.RESET_ALL,
            Fore.BLUE,
            display_config_path,
            Style.RESET_ALL,
        )

    duplicate_merge_conflicts: list[str] = []
    duplicate_merge_changes = 0
    duplicate_merge_details: list[tuple[str, int]] = []

    if merge_duplicates:
        (
            raw_config,
            duplicate_merge_conflicts,
            duplicate_merge_changes,
            duplicate_merge_details,
        ) = merge_duplicate_workspace_roots(raw_config, duplicate_root_occurrences)
        for message in duplicate_merge_conflicts:
            log.warning(message)
        if duplicate_merge_changes and duplicate_merge_details:
            for label, occurrence_count in duplicate_merge_details:
                log.info(
                    "%s•%s Merged %s%d%s duplicate entr%s for workspace root %s%s%s",
                    Fore.BLUE,
                    Style.RESET_ALL,
                    Fore.YELLOW,
                    occurrence_count,
                    Style.RESET_ALL,
                    "y" if occurrence_count == 1 else "ies",
                    Fore.MAGENTA,
                    label,
                    Style.RESET_ALL,
                )
    elif duplicate_root_occurrences:
        duplicate_merge_details = [
            (label, len(values)) for label, values in duplicate_root_occurrences.items()
        ]
        for label, occurrence_count in duplicate_merge_details:
            log.warning(
                "%s•%s Duplicate workspace root %s%s%s appears %s%d%s time%s; "
                "skipping merge because --no-merge was provided.",
                Fore.BLUE,
                Style.RESET_ALL,
                Fore.MAGENTA,
                label,
                Style.RESET_ALL,
                Fore.YELLOW,
                occurrence_count,
                Style.RESET_ALL,
                "" if occurrence_count == 1 else "s",
            )

    explicit_relative_override = workspace_root_override in {".", "./"}
    preserve_cwd_label = explicit_relative_override or allow_relative_workspace

    if merge_duplicates:
        (
            raw_config,
            workspace_map,
            merge_conflicts,
            merge_changes,
        ) = normalize_workspace_roots(
            raw_config,
            cwd=cwd,
            home=home,
            preserve_cwd_label=preserve_cwd_label,
        )
    else:
        (
            _,
            workspace_map,
            merge_conflicts,
            merge_changes,
        ) = normalize_workspace_roots(
            raw_config,
            cwd=cwd,
            home=home,
            preserve_cwd_label=preserve_cwd_label,
        )

    for message in merge_conflicts:
        log.warning(message)

    found_repos: list[tuple[str, str, pathlib.Path]] = []

    override_workspace_path: pathlib.Path | None = None
    if workspace_root_override:
        override_workspace_path = _resolve_workspace_path(
            workspace_root_override,
            None,
            cwd=cwd,
        )

    if recursive:
        for root, dirs, _ in os.walk(scan_dir):
            if ".git" in dirs:
                repo_path = pathlib.Path(root)
                repo_name = repo_path.name
                repo_url = get_git_origin_url(repo_path)

                if not repo_url:
                    log.warning(
                        "Could not determine remote URL for git repository "
                        "at %s. Skipping.",
                        PrivatePath(repo_path),
                    )
                    continue

                workspace_path = override_workspace_path or scan_dir
                found_repos.append((repo_name, repo_url, workspace_path))
    else:
        for item in scan_dir.iterdir():
            if item.is_dir() and (item / ".git").is_dir():
                repo_name = item.name
                repo_url = get_git_origin_url(item)

                if not repo_url:
                    log.warning(
                        "Could not determine remote URL for git repository "
                        "at %s. Skipping.",
                        PrivatePath(item),
                    )
                    continue

                workspace_path = override_workspace_path or scan_dir
                found_repos.append((repo_name, repo_url, workspace_path))

    if not found_repos:
        log.info(
            "%s!%s No git repositories found in %s%s%s. Nothing to import.",
            Fore.YELLOW,
            Style.RESET_ALL,
            Fore.BLUE,
            PrivatePath(scan_dir),
            Style.RESET_ALL,
        )
        return

    repos_to_add: list[tuple[str, str, pathlib.Path]] = []
    existing_repos: list[tuple[str, str, pathlib.Path]] = []

    for name, url, workspace_path in found_repos:
        workspace_label = workspace_map.get(workspace_path)
        if workspace_label is None:
            workspace_label = workspace_root_label(
                workspace_path,
                cwd=cwd,
                home=home,
                preserve_cwd_label=preserve_cwd_label,
            )
            workspace_map[workspace_path] = workspace_label
            raw_config.setdefault(workspace_label, {})

        target_section = raw_config.get(workspace_label, {})
        if isinstance(target_section, dict) and name in target_section:
            existing_repos.append((name, url, workspace_path))
        else:
            repos_to_add.append((name, url, workspace_path))

    if existing_repos:
        # Show summary only when there are many existing repos
        if len(existing_repos) > 5:
            log.info(
                "%s!%s Found %s%d%s existing repositories already in configuration.",
                Fore.YELLOW,
                Style.RESET_ALL,
                Fore.CYAN,
                len(existing_repos),
                Style.RESET_ALL,
            )
        else:
            # Show details only for small numbers
            log.info(
                "%s!%s Found %s%d%s existing repositories in configuration:",
                Fore.YELLOW,
                Style.RESET_ALL,
                Fore.CYAN,
                len(existing_repos),
                Style.RESET_ALL,
            )
            for name, url, workspace_path in existing_repos:
                workspace_label = workspace_map.get(workspace_path)
                if workspace_label is None:
                    workspace_label = workspace_root_label(
                        workspace_path,
                        cwd=cwd,
                        home=home,
                        preserve_cwd_label=preserve_cwd_label,
                    )
                    workspace_map[workspace_path] = workspace_label
                    raw_config.setdefault(workspace_label, {})
                log.info(
                    "  %s•%s %s%s%s (%s%s%s) at %s%s%s%s in %s%s%s",
                    Fore.BLUE,
                    Style.RESET_ALL,
                    Fore.CYAN,
                    name,
                    Style.RESET_ALL,
                    Fore.YELLOW,
                    url,
                    Style.RESET_ALL,
                    Fore.MAGENTA,
                    workspace_label,
                    name,
                    Style.RESET_ALL,
                    Fore.BLUE,
                    display_config_path,
                    Style.RESET_ALL,
                )

    changes_made = merge_duplicates and (
        merge_changes > 0 or duplicate_merge_changes > 0
    )

    if not repos_to_add:
        if existing_repos:
            log.info(
                "%s✓%s All found repositories already exist in the configuration. "
                "%sNothing to do.%s",
                Fore.GREEN,
                Style.RESET_ALL,
                Fore.GREEN,
                Style.RESET_ALL,
            )
        if changes_made and not dry_run:
            try:
                save_config_yaml(config_file_path, raw_config)
                log.info(
                    "%s✓%s Successfully updated %s%s%s.",
                    Fore.GREEN,
                    Style.RESET_ALL,
                    Fore.BLUE,
                    display_config_path,
                    Style.RESET_ALL,
                )
            except Exception:
                log.exception(
                    "Error saving config to %s",
                    PrivatePath(config_file_path),
                )
                if log.isEnabledFor(logging.DEBUG):
                    traceback.print_exc()
            return
        return

    # Show what will be added
    log.info(
        "\n%sFound %d new %s to %s:%s",
        Fore.GREEN,
        len(repos_to_add),
        "repository" if len(repos_to_add) == 1 else "repositories",
        "preview" if dry_run else "import",
        Style.RESET_ALL,
    )
    for repo_name, repo_url, _determined_base_key in repos_to_add:
        log.info(
            "  %s+%s %s%s%s (%s%s%s)",
            Fore.GREEN,
            Style.RESET_ALL,
            Fore.CYAN,
            repo_name,
            Style.RESET_ALL,
            Fore.YELLOW,
            repo_url,
            Style.RESET_ALL,
        )

    if dry_run:
        log.info(
            "\n%s→%s Dry run complete. No changes made to %s%s%s.",
            Fore.YELLOW,
            Style.RESET_ALL,
            Fore.BLUE,
            display_config_path,
            Style.RESET_ALL,
        )
        return

    if not yes:
        confirm = input(
            f"\n{Fore.CYAN}Import these repositories? [y/N]: {Style.RESET_ALL}",
        ).lower()
        if confirm not in {"y", "yes"}:
            log.info("%s✗%s Aborted by user.", Fore.RED, Style.RESET_ALL)
            return

    for repo_name, repo_url, workspace_path in repos_to_add:
        workspace_label = workspace_map.get(workspace_path)
        if workspace_label is None:
            workspace_label = workspace_root_label(
                workspace_path,
                cwd=cwd,
                home=home,
                preserve_cwd_label=preserve_cwd_label,
            )
            workspace_map[workspace_path] = workspace_label

        if workspace_label not in raw_config:
            raw_config[workspace_label] = {}
        elif not isinstance(raw_config[workspace_label], dict):
            log.warning(
                "Workspace root '%s' in config is not a dictionary. Skipping repo %s.",
                workspace_label,
                repo_name,
            )
            continue

        if repo_name not in raw_config[workspace_label]:
            raw_config[workspace_label][repo_name] = {"repo": repo_url}
            log.info(
                "%s+%s Importing %s'%s'%s (%s%s%s) under '%s%s%s'.",
                Fore.GREEN,
                Style.RESET_ALL,
                Fore.CYAN,
                repo_name,
                Style.RESET_ALL,
                Fore.YELLOW,
                repo_url,
                Style.RESET_ALL,
                Fore.MAGENTA,
                workspace_label,
                Style.RESET_ALL,
            )
            changes_made = True

    if changes_made:
        try:
            save_config_yaml(config_file_path, raw_config)
            log.info(
                "%s✓%s Successfully updated %s%s%s.",
                Fore.GREEN,
                Style.RESET_ALL,
                Fore.BLUE,
                display_config_path,
                Style.RESET_ALL,
            )
        except Exception:
            log.exception(
                "Error saving config to %s",
                PrivatePath(config_file_path),
            )
            if log.isEnabledFor(logging.DEBUG):
                traceback.print_exc()
            return
    else:
        log.info(
            "%s✓%s No changes made to the configuration.",
            Fore.GREEN,
            Style.RESET_ALL,
        )



from __future__ import annotations

import copy
import json
import pathlib
import typing as t

import yaml

FormatLiteral = t.Literal["json", "yaml"]
RawConfigData: t.TypeAlias = dict[t.Any, t.Any]


class ConfigReader:
    r"""Parse string data (YAML and JSON) into a dictionary.

    >>> cfg = ConfigReader({ "session_name": "my session" })
    >>> cfg.dump("yaml")
    'session_name: my session\n'
    >>> cfg.dump("json")
    '{\n  "session_name": "my session"\n}'
    """

    def __init__(self, content: RawConfigData) -> None:
        self.content = content

    @staticmethod
    def _load(fmt: FormatLiteral, content: str) -> dict[str, t.Any]:
        """Load raw config data and directly return it.

        >>> ConfigReader._load("json", '{ "session_name": "my session" }')
        {'session_name': 'my session'}

        >>> ConfigReader._load("yaml", 'session_name: my session')
        {'session_name': 'my session'}
        """
        if fmt == "yaml":
            return t.cast(
                "dict[str, t.Any]",
                yaml.load(
                    content,
                    Loader=yaml.SafeLoader,
                ),
            )
        if fmt == "json":
            return t.cast("dict[str, t.Any]", json.loads(content))
        msg = f"{fmt} not supported in configuration"
        raise NotImplementedError(msg)

    @classmethod
    def load(cls, fmt: FormatLiteral, content: str) -> ConfigReader:
        """Load raw config data into a ConfigReader instance (to dump later).

        >>> cfg = ConfigReader.load("json", '{ "session_name": "my session" }')
        >>> cfg
        <tmuxp.config_reader.ConfigReader object at ...>
        >>> cfg.content
        {'session_name': 'my session'}

        >>> cfg = ConfigReader.load("yaml", 'session_name: my session')
        >>> cfg
        <tmuxp.config_reader.ConfigReader object at ...>
        >>> cfg.content
        {'session_name': 'my session'}
        """
        return cls(
            content=cls._load(
                fmt=fmt,
                content=content,
            ),
        )

    @classmethod
    def _from_file(cls, path: pathlib.Path) -> dict[str, t.Any]:
        r"""Load data from file path directly to dictionary.

        **YAML file**

        *For demonstration only,* create a YAML file:

        >>> yaml_file = tmp_path / 'my_config.yaml'
        >>> yaml_file.write_text('session_name: my session', encoding='utf-8')
        24

        *Read YAML file*:

        >>> ConfigReader._from_file(yaml_file)
        {'session_name': 'my session'}

        **JSON file**

        *For demonstration only,* create a JSON file:

        >>> json_file = tmp_path / 'my_config.json'
        >>> json_file.write_text('{"session_name": "my session"}', encoding='utf-8')
        30

        *Read JSON file*:

        >>> ConfigReader._from_file(json_file)
        {'session_name': 'my session'}
        """
        assert isinstance(path, pathlib.Path)
        content = path.open(encoding="utf-8").read()

        # TODO(#?): Align this loader with the duplicate-aware YAML handling that
        # ``vcspull fmt`` introduced in November 2025. The formatter now uses a
        # custom SafeLoader subclass to retain and merge duplicate workspace root
        # sections so repos are never overwritten. ConfigReader currently drops
        # later duplicates because PyYAML keeps only the last key. Options:
        # 1) Extract the formatter's loader/merge helpers into a shared utility
        #    that ConfigReader can reuse here;
        # 2) Replace ConfigReader entirely when reading vcspull configs and call
        #    the formatter helpers directly;
        # 3) Keep this basic loader but add an opt-in path for duplicate-aware
        #    parsing so commands like ``vcspull add`` can avoid data loss.
        # Revisit once the new ``vcspull add`` flow lands so both commands share
        # the same duplication safeguards.

        if path.suffix in {".yaml", ".yml"}:
            fmt: FormatLiteral = "yaml"
        elif path.suffix == ".json":
            fmt = "json"
        else:
            msg = f"{path.suffix} not supported in {path}"
            raise NotImplementedError(msg)

        return cls._load(
            fmt=fmt,
            content=content,
        )

    @classmethod
    def from_file(cls, path: pathlib.Path) -> ConfigReader:
        r"""Load data from file path.

        **YAML file**

        *For demonstration only,* create a YAML file:

        >>> yaml_file = tmp_path / 'my_config.yaml'
        >>> yaml_file.write_text('session_name: my session', encoding='utf-8')
        24

        *Read YAML file*:

        >>> cfg = ConfigReader.from_file(yaml_file)
        >>> cfg
        <tmuxp.config_reader.ConfigReader object at ...>

        >>> cfg.content
        {'session_name': 'my session'}

        **JSON file**

        *For demonstration only,* create a JSON file:

        >>> json_file = tmp_path / 'my_config.json'
        >>> json_file.write_text('{"session_name": "my session"}', encoding='utf-8')
        30

        *Read JSON file*:

        >>> cfg = ConfigReader.from_file(json_file)
        >>> cfg
        <tmuxp.config_reader.ConfigReader object at ...>

        >>> cfg.content
        {'session_name': 'my session'}
        """
        return cls(content=cls._from_file(path=path))

    @staticmethod
    def _dump(
        fmt: FormatLiteral,
        content: RawConfigData,
        indent: int = 2,
        **kwargs: t.Any,
    ) -> str:
        r"""Dump directly.

        >>> ConfigReader._dump("yaml", { "session_name": "my session" })
        'session_name: my session\n'

        >>> ConfigReader._dump("json", { "session_name": "my session" })
        '{\n  "session_name": "my session"\n}'
        """
        if fmt == "yaml":
            return yaml.dump(
                content,
                indent=2,
                default_flow_style=False,
                Dumper=yaml.SafeDumper,
            )
        if fmt == "json":
            return json.dumps(
                content,
                indent=2,
            )
        msg = f"{fmt} not supported in config"
        raise NotImplementedError(msg)

    def dump(self, fmt: FormatLiteral, indent: int = 2, **kwargs: t.Any) -> str:
        r"""Dump via ConfigReader instance.

        >>> cfg = ConfigReader({ "session_name": "my session" })
        >>> cfg.dump("yaml")
        'session_name: my session\n'
        >>> cfg.dump("json")
        '{\n  "session_name": "my session"\n}'
        """
        return self._dump(
            fmt=fmt,
            content=self.content,
            indent=indent,
            **kwargs,
        )


class _DuplicateTrackingSafeLoader(yaml.SafeLoader):
    """SafeLoader that records duplicate top-level keys."""

    def __init__(self, stream: str) -> None:
        super().__init__(stream)
        self.top_level_key_values: dict[t.Any, list[t.Any]] = {}
        self._mapping_depth = 0
        self.top_level_items: list[tuple[t.Any, t.Any]] = []


def _duplicate_tracking_construct_mapping(
    loader: _DuplicateTrackingSafeLoader,
    node: yaml.nodes.MappingNode,
    deep: bool = False,
) -> dict[t.Any, t.Any]:
    loader._mapping_depth += 1
    loader.flatten_mapping(node)
    mapping: dict[t.Any, t.Any] = {}

    for key_node, value_node in node.value:
        construct = t.cast(
            "t.Callable[[yaml.nodes.Node], t.Any]",
            loader.construct_object,
        )
        key = construct(key_node)
        value = construct(value_node)

        if loader._mapping_depth == 1:
            duplicated_value = copy.deepcopy(value)
            loader.top_level_key_values.setdefault(key, []).append(duplicated_value)
            loader.top_level_items.append((copy.deepcopy(key), duplicated_value))

        mapping[key] = value

    loader._mapping_depth -= 1
    return mapping


_DuplicateTrackingSafeLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    _duplicate_tracking_construct_mapping,
)


class DuplicateAwareConfigReader(ConfigReader):
    """ConfigReader that tracks duplicate top-level YAML sections."""

    def __init__(
        self,
        content: RawConfigData,
        *,
        duplicate_sections: dict[str, list[t.Any]] | None = None,
        top_level_items: list[tuple[str, t.Any]] | None = None,
    ) -> None:
        super().__init__(content)
        self._duplicate_sections = duplicate_sections or {}
        self._top_level_items = top_level_items or []

    @property
    def duplicate_sections(self) -> dict[str, list[t.Any]]:
        """Mapping of top-level keys to the list of duplicated values."""
        return self._duplicate_sections

    @property
    def top_level_items(self) -> list[tuple[str, t.Any]]:
        """Ordered list of top-level items, including duplicates."""
        return copy.deepcopy(self._top_level_items)

    @classmethod
    def _load_yaml_with_duplicates(
        cls,
        content: str,
    ) -> tuple[dict[str, t.Any], dict[str, list[t.Any]], list[tuple[str, t.Any]]]:
        loader = _DuplicateTrackingSafeLoader(content)

        try:
            data = loader.get_single_data()
        finally:
            dispose = t.cast("t.Callable[[], None]", loader.dispose)
            dispose()

        if data is None:
            loaded: dict[str, t.Any] = {}
        else:
            if not isinstance(data, dict):
                msg = "Loaded configuration is not a mapping"
                raise TypeError(msg)
            loaded = t.cast("dict[str, t.Any]", data)

        duplicate_sections = {
            t.cast("str", key): values
            for key, values in loader.top_level_key_values.items()
            if len(values) > 1
        }

        top_level_items = [
            (t.cast("str", key), copy.deepcopy(value))
            for key, value in loader.top_level_items
        ]

        return loaded, duplicate_sections, top_level_items

    @classmethod
    def _load_from_path(
        cls,
        path: pathlib.Path,
    ) -> tuple[dict[str, t.Any], dict[str, list[t.Any]], list[tuple[str, t.Any]]]:
        if path.suffix.lower() in {".yaml", ".yml"}:
            content = path.read_text(encoding="utf-8")
            return cls._load_yaml_with_duplicates(content)

        return ConfigReader._from_file(path), {}, []

    @classmethod
    def from_file(cls, path: pathlib.Path) -> DuplicateAwareConfigReader:
        content, duplicate_sections, top_level_items = cls._load_from_path(path)
        return cls(
            content,
            duplicate_sections=duplicate_sections,
            top_level_items=top_level_items,
        )

    @classmethod
    def _from_file(cls, path: pathlib.Path) -> dict[str, t.Any]:
        content, _, _ = cls._load_from_path(path)
        return content

    @classmethod
    def load_with_duplicates(
        cls,
        path: pathlib.Path,
    ) -> tuple[dict[str, t.Any], dict[str, list[t.Any]], list[tuple[str, t.Any]]]:
        reader = cls.from_file(path)
        return reader.content, reader.duplicate_sections, reader.top_level_items



"""Tests for sync planner helper utilities."""

from __future__ import annotations

import subprocess
import typing as t

import pytest

from vcspull.cli._output import PlanAction
from vcspull.cli.sync import SyncPlanConfig, _determine_plan_action, _maybe_fetch

if t.TYPE_CHECKING:
    import pathlib


class MaybeFetchFixture(t.NamedTuple):
    """Fixture for _maybe_fetch behaviours."""

    test_id: str
    fetch: bool
    offline: bool
    create_repo: bool
    create_git_dir: bool
    subprocess_behavior: str | None
    expected_result: tuple[bool, str | None]


MAYBE_FETCH_FIXTURES: list[MaybeFetchFixture] = [
    MaybeFetchFixture(
        test_id="offline-short-circuit",
        fetch=True,
        offline=True,
        create_repo=True,
        create_git_dir=True,
        subprocess_behavior=None,
        expected_result=(True, None),
    ),
    MaybeFetchFixture(
        test_id="no-git-directory",
        fetch=True,
        offline=False,
        create_repo=True,
        create_git_dir=False,
        subprocess_behavior=None,
        expected_result=(True, None),
    ),
    MaybeFetchFixture(
        test_id="missing-git-executable",
        fetch=True,
        offline=False,
        create_repo=True,
        create_git_dir=True,
        subprocess_behavior="file-not-found",
        expected_result=(False, "git executable not found"),
    ),
    MaybeFetchFixture(
        test_id="fetch-non-zero-exit",
        fetch=True,
        offline=False,
        create_repo=True,
        create_git_dir=True,
        subprocess_behavior="non-zero",
        expected_result=(False, "remote rejected"),
    ),
    MaybeFetchFixture(
        test_id="fetch-oserror",
        fetch=True,
        offline=False,
        create_repo=True,
        create_git_dir=True,
        subprocess_behavior="os-error",
        expected_result=(False, "Permission denied"),
    ),
    MaybeFetchFixture(
        test_id="fetch-disabled",
        fetch=False,
        offline=False,
        create_repo=True,
        create_git_dir=True,
        subprocess_behavior="non-zero",
        expected_result=(True, None),
    ),
]


@pytest.mark.parametrize(
    list(MaybeFetchFixture._fields),
    MAYBE_FETCH_FIXTURES,
    ids=[fixture.test_id for fixture in MAYBE_FETCH_FIXTURES],
)
def test_maybe_fetch_behaviour(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    test_id: str,
    fetch: bool,
    offline: bool,
    create_repo: bool,
    create_git_dir: bool,
    subprocess_behavior: str | None,
    expected_result: tuple[bool, str | None],
) -> None:
    """Ensure _maybe_fetch handles subprocess outcomes correctly."""
    repo_path = tmp_path / "repo"
    if create_repo:
        repo_path.mkdir()
    if create_git_dir:
        (repo_path / ".git").mkdir(parents=True, exist_ok=True)

    if subprocess_behavior:

        def _patched_run(
            *args: t.Any,
            **kwargs: t.Any,
        ) -> subprocess.CompletedProcess[str]:
            if subprocess_behavior == "file-not-found":
                error_message = "git executable not found"
                raise FileNotFoundError(error_message)
            if subprocess_behavior == "os-error":
                error_message = "Permission denied"
                raise OSError(error_message)
            if subprocess_behavior == "non-zero":
                return subprocess.CompletedProcess(
                    args=args[0],
                    returncode=1,
                    stdout="",
                    stderr="remote rejected",
                )
            return subprocess.CompletedProcess(
                args=args[0],
                returncode=0,
                stdout="",
                stderr="",
            )

        monkeypatch.setattr("subprocess.run", _patched_run)

    result = _maybe_fetch(
        repo_path=repo_path,
        config=SyncPlanConfig(fetch=fetch, offline=offline),
    )

    assert result == expected_result


class DeterminePlanActionFixture(t.NamedTuple):
    """Fixture for _determine_plan_action outcomes."""

    test_id: str
    status: dict[str, t.Any]
    config: SyncPlanConfig
    expected_action: PlanAction
    expected_detail: str


DETERMINE_PLAN_ACTION_FIXTURES: list[DeterminePlanActionFixture] = [
    DeterminePlanActionFixture(
        test_id="missing-repo",
        status={"exists": False},
        config=SyncPlanConfig(fetch=False, offline=False),
        expected_action=PlanAction.CLONE,
        expected_detail="missing",
    ),
    DeterminePlanActionFixture(
        test_id="not-git",
        status={"exists": True, "is_git": False},
        config=SyncPlanConfig(fetch=True, offline=False),
        expected_action=PlanAction.BLOCKED,
        expected_detail="not a git repository",
    ),
    DeterminePlanActionFixture(
        test_id="dirty-working-tree",
        status={"exists": True, "is_git": True, "clean": False},
        config=SyncPlanConfig(fetch=True, offline=False),
        expected_action=PlanAction.BLOCKED,
        expected_detail="working tree has local changes",
    ),
    DeterminePlanActionFixture(
        test_id="diverged",
        status={"exists": True, "is_git": True, "clean": True, "ahead": 2, "behind": 3},
        config=SyncPlanConfig(fetch=True, offline=False),
        expected_action=PlanAction.BLOCKED,
        expected_detail="diverged (ahead 2, behind 3)",
    ),
    DeterminePlanActionFixture(
        test_id="behind-remote",
        status={"exists": True, "is_git": True, "clean": True, "ahead": 0, "behind": 4},
        config=SyncPlanConfig(fetch=True, offline=False),
        expected_action=PlanAction.UPDATE,
        expected_detail="behind 4",
    ),
    DeterminePlanActionFixture(
        test_id="ahead-remote",
        status={"exists": True, "is_git": True, "clean": True, "ahead": 1, "behind": 0},
        config=SyncPlanConfig(fetch=True, offline=False),
        expected_action=PlanAction.BLOCKED,
        expected_detail="ahead by 1",
    ),
    DeterminePlanActionFixture(
        test_id="up-to-date",
        status={"exists": True, "is_git": True, "clean": True, "ahead": 0, "behind": 0},
        config=SyncPlanConfig(fetch=True, offline=False),
        expected_action=PlanAction.UNCHANGED,
        expected_detail="up to date",
    ),
    DeterminePlanActionFixture(
        test_id="offline-remote-unknown",
        status={
            "exists": True,
            "is_git": True,
            "clean": True,
            "ahead": None,
            "behind": None,
        },
        config=SyncPlanConfig(fetch=True, offline=True),
        expected_action=PlanAction.UPDATE,
        expected_detail="remote state unknown (offline)",
    ),
    DeterminePlanActionFixture(
        test_id="needs-fetch",
        status={
            "exists": True,
            "is_git": True,
            "clean": True,
            "ahead": None,
            "behind": None,
        },
        config=SyncPlanConfig(fetch=True, offline=False),
        expected_action=PlanAction.UPDATE,
        expected_detail="remote state unknown; use --fetch",
    ),
]


@pytest.mark.parametrize(
    list(DeterminePlanActionFixture._fields),
    DETERMINE_PLAN_ACTION_FIXTURES,
    ids=[fixture.test_id for fixture in DETERMINE_PLAN_ACTION_FIXTURES],
)
def test_determine_plan_action(
    test_id: str,
    status: dict[str, t.Any],
    config: SyncPlanConfig,
    expected_action: PlanAction,
    expected_detail: str,
) -> None:
    """Verify _determine_plan_action handles edge cases."""
    action, detail = _determine_plan_action(status, config=config)
    assert action is expected_action
    assert detail == expected_detail



from __future__ import annotations

import os
from pathlib import Path

from vcspull._internal.private_path import PrivatePath


def test_str_collapses_home_directory() -> None:
    home = Path.home()
    assert str(PrivatePath(home)) == "~"

    project_path = PrivatePath(home / "projects" / "vcspull")
    assert str(project_path) == f"~{os.sep}projects{os.sep}vcspull"


def test_repr_uses_tilde_for_home_directory() -> None:
    home = Path.home()
    project_path = PrivatePath(home / "projects")
    assert repr(project_path) == f"PrivatePath('~{os.sep}projects')"


def test_absolute_paths_outside_home_are_unmodified() -> None:
    path = PrivatePath("/tmp/vcspull-example")
    assert str(path) == "/tmp/vcspull-example"
    assert repr(path) == "PrivatePath('/tmp/vcspull-example')"


def test_existing_tilde_prefix_is_preserved() -> None:
    path = PrivatePath("~/.config/vcspull")
    assert str(path) == "~/.config/vcspull"


def test_trailing_slash_collapses_to_directory_label() -> None:
    home = Path.home()
    path = PrivatePath(home / "projects/")
    assert str(path) == f"~{os.sep}projects"


def test_relative_path_unmodified() -> None:
    path = PrivatePath("relative/path")
    assert str(path) == "relative/path"
